"""Module provides generic interface to running Galaxy tools and workflows."""

import os
import sys
import tempfile
import time
import traceback
from datetime import datetime
from urllib.parse import urljoin

import bioblend
from bioblend.util import attach_file

try:
    from galaxy.tool_util.client.staging import StagingInterface
except ImportError:
    from galaxy.tool_util.client.staging import StagingInterace as StagingInterface
from galaxy.tool_util.cwl.util import (
    invocation_to_output,
    output_to_cwl_json,
    tool_response_to_output,
)
from galaxy.tool_util.parser import get_tool_source
from galaxy.util import (
    safe_makedirs,
    unicodify,
)
from requests.exceptions import (
    HTTPError,
    RequestException,
)

from planemo.galaxy.api import summarize_history
from planemo.io import wait_on
from planemo.runnable import (
    ErrorRunResponse,
    get_outputs,
    RunnableType,
    SuccessfulRunResponse,
)

DEFAULT_HISTORY_NAME = "CWL Target History"
ERR_NO_SUCH_TOOL = (
    "Failed to find tool with ID [%s] in Galaxy - cannot execute job. "
    "You may need to enable verbose logging and determine why the tool did not load. [%s]"
)


def execute(ctx, config, runnable, job_path, **kwds):
    """Execute a Galaxy activity."""
    try:
        start_datetime = datetime.now()
        return _execute(ctx, config, runnable, job_path, **kwds)
    except Exception as e:
        end_datetime = datetime.now()
        ctx.log("Failed to execute Galaxy activity, throwing ErrorRunResponse")
        traceback.print_exc(file=sys.stdout)
        return ErrorRunResponse(unicodify(e), start_datetime=start_datetime, end_datetime=end_datetime)


def _verified_tool_id(runnable, user_gi):
    tool_id = _tool_id(runnable.path)
    try:
        user_gi.tools.show_tool(tool_id)
    except Exception as e:
        raise Exception(ERR_NO_SUCH_TOOL % (tool_id, e))
    return tool_id


def _inputs_representation(runnable):
    if runnable.type == RunnableType.cwl_tool:
        inputs_representation = "cwl"
    else:
        inputs_representation = "galaxy"
    return inputs_representation


def log_contents_str(config):
    if hasattr(config, "log_contents"):
        return config.log_contents
    else:
        return "No log for this engine type."


class PlanemoStagingInterface(StagingInterface):
    def __init__(self, ctx, runnable, user_gi, version_major, simultaneous_uploads):
        self._ctx = ctx
        self._user_gi = user_gi
        self._runnable = runnable
        self._version_major = version_major
        self._simultaneous_uploads = simultaneous_uploads

    def _post(self, api_path, payload, files_attached=False):
        url = urljoin(self._user_gi.url, "api/" + api_path)
        if payload.get("__files"):  # put attached files where BioBlend expects them
            files_attached = True
            for k, v in payload["__files"].items():
                payload[k] = v
            del payload["__files"]
        return self._user_gi.make_post_request(url, payload=payload, files_attached=files_attached)

    def _attach_file(self, path):
        return attach_file(path)

    def _handle_job(self, job_response):
        if not self._simultaneous_uploads:
            job_id = job_response["id"]
            _wait_for_job(self._user_gi, job_id)

    @property
    def use_fetch_api(self):
        # hack around this not working for galaxy_tools - why is that :(
        return self._runnable.type != RunnableType.galaxy_tool and self._version_major >= "20.09"

    # extension point for planemo to override logging
    def _log(self, message):
        self._ctx.vlog(message)


def _execute(ctx, config, runnable, job_path, **kwds):  # noqa C901
    user_gi = config.user_gi
    admin_gi = config.gi

    start_datetime = datetime.now()
    try:
        job_dict, _, history_id = stage_in(ctx, runnable, config, job_path, **kwds)
    except Exception:
        ctx.vlog("Problem with staging in data for Galaxy activities...")
        raise
    if runnable.type in [RunnableType.galaxy_tool, RunnableType.cwl_tool]:
        response_class = GalaxyToolRunResponse
        tool_id = _verified_tool_id(runnable, user_gi)
        inputs_representation = _inputs_representation(runnable)
        run_tool_payload = dict(
            history_id=history_id,
            tool_id=tool_id,
            inputs=job_dict,
            inputs_representation=inputs_representation,
        )
        ctx.vlog("Post to Galaxy tool API with payload [%s]" % run_tool_payload)
        tool_run_response = user_gi.tools._post(run_tool_payload)

        if not kwds.get("no_wait"):
            job = tool_run_response["jobs"][0]
            job_id = job["id"]
            try:
                final_state = _wait_for_job(user_gi, job_id, timeout=kwds.get("test_timeout"))
            except Exception:
                summarize_history(ctx, user_gi, history_id)
                raise
            if final_state != "ok":
                msg = "Failed to run CWL tool job final job state is [%s]." % final_state
                summarize_history(ctx, user_gi, history_id)
                raise Exception(msg)

            ctx.vlog("Final job state was ok, fetching details for job [%s]" % job_id)
            job_info = admin_gi.jobs.show_job(job_id)
            response_kwds = {
                "job_info": job_info,
                "api_run_response": tool_run_response,
            }
            if ctx.verbose:
                summarize_history(ctx, user_gi, history_id)

    elif runnable.type in [RunnableType.galaxy_workflow, RunnableType.cwl_workflow]:
        response_class = GalaxyWorkflowRunResponse
        workflow_id = config.workflow_id_for_runnable(runnable)
        ctx.vlog(f"Found Galaxy workflow ID [{workflow_id}] for URI [{runnable.uri}]")
        invocation = user_gi.workflows.invoke_workflow(
            workflow_id,
            inputs=job_dict,
            history_id=history_id,
            allow_tool_state_corrections=True,
            inputs_by="name",
        )
        invocation_id = invocation["id"]

        ctx.vlog("Waiting for invocation [%s]" % invocation_id)
        polling_backoff = kwds.get("polling_backoff", 0)
        final_invocation_state = "new"
        error_message = ""
        try:
            final_invocation_state = _wait_for_invocation(
                ctx, user_gi, history_id, workflow_id, invocation_id, polling_backoff
            )
            assert final_invocation_state == "scheduled"
        except Exception:
            ctx.vlog("Problem waiting on invocation...")
            summarize_history(ctx, user_gi, history_id)
            error_message = "Final invocation state is [%s]" % final_invocation_state
        ctx.vlog("Final invocation state is [%s]" % final_invocation_state)

        if not kwds.get("no_wait"):
            final_state = _wait_for_invocation_jobs(ctx, user_gi, invocation_id, polling_backoff)
            if final_state != "ok":
                msg = "Failed to run workflow final history state is [%s]." % final_state
                error_message = msg if not error_message else f"{error_message}. {msg}"
                ctx.vlog(msg)
                summarize_history(ctx, user_gi, history_id)
            else:
                ctx.vlog("Final history state is 'ok'")

        response_kwds = {
            "workflow_id": workflow_id,
            "invocation_id": invocation_id,
            "history_state": final_state if not kwds.get("no_wait") else None,
            "invocation_state": final_invocation_state,
            "error_message": error_message,
        }

    else:
        raise NotImplementedError()

    run_response = response_class(
        ctx=ctx,
        runnable=runnable,
        user_gi=user_gi,
        history_id=history_id,
        log=log_contents_str(config),
        start_datetime=start_datetime,
        end_datetime=datetime.now(),
        **response_kwds,
    )
    if kwds.get("download_outputs", True):
        output_directory = kwds.get("output_directory", None)
        ctx.vlog("collecting outputs from run...")
        run_response.collect_outputs(ctx, output_directory)
        ctx.vlog("collecting outputs complete")
    return run_response


def stage_in(ctx, runnable, config, job_path, **kwds):  # noqa C901
    # only upload objects as files/collections for CWL workflows...
    tool_or_workflow = "tool" if runnable.type != RunnableType.cwl_workflow else "workflow"
    to_posix_lines = runnable.type.is_galaxy_artifact
    simultaneous_uploads = kwds.get("simultaneous_uploads", False)
    user_gi = config.user_gi
    history_id = _history_id(user_gi, **kwds)
    psi = PlanemoStagingInterface(ctx, runnable, user_gi, config.version_major, simultaneous_uploads)
    job_dict, datasets = psi.stage(
        tool_or_workflow,
        history_id=history_id,
        job_path=job_path,
        use_path_paste=config.use_path_paste,
        to_posix_lines=to_posix_lines,
    )

    if datasets and kwds.get("check_uploads_ok", True):
        ctx.vlog("uploaded datasets [%s] for activity, checking history state" % datasets)
        final_state = _wait_for_history(ctx, user_gi, history_id)

        for (dataset, path) in datasets:
            dataset_details = user_gi.histories.show_dataset(
                history_id,
                dataset["id"],
            )
            ctx.vlog(f"Uploaded dataset for path [{path}] with metadata [{dataset_details}]")
    else:
        # Mark uploads as ok because nothing to do.
        final_state = "ok"

    ctx.vlog("final state is %s" % final_state)
    if final_state != "ok":
        msg = "Failed to upload data, upload state is [%s]." % final_state
        summarize_history(ctx, user_gi, history_id)
        raise Exception(msg)
    return job_dict, datasets, history_id


def _file_path_to_name(file_path):
    if file_path is not None:
        name = os.path.basename(file_path)
    else:
        name = "defaultname"
    return name


def execute_rerun(ctx, config, rerunnable, **kwds):
    rerun_successful = True
    user_gi = config.user_gi
    if rerunnable.rerunnable_type == "history":
        job_ids = [job["id"] for job in user_gi.jobs.get_jobs(history_id=rerunnable.rerunnable_id, state="error")]
    elif rerunnable.rerunnable_type == "invocation":
        job_ids = [job["id"] for job in user_gi.jobs.get_jobs(invocation_id=rerunnable.rerunnable_id, state="error")]
    elif rerunnable.rerunnable_type == "job":
        job_ids = [rerunnable.rerunnable_id]
    # elif rerunnable.rerunnable_type = 'collection':
    else:
        raise Exception(f"Unknown Rerunnable type {rerunnable.rerunnable_type}")

    for job_id in job_ids:
        try:
            user_gi.jobs.rerun_job(job_id, remap=True)
        except (ValueError, bioblend.ConnectionError):
            rerun_successful = False
            if rerunnable.rerunnable_type == "job":
                ctx.log(f"Job {job_id} could not be rerun with dataset remapping.")
            else:
                ctx.log(
                    f"Job {job_id} associated with {rerunnable.rerunnable_type} {rerunnable.rerunnable_id} "
                    "could not be rerun with dataset remapping."
                )
        else:
            if rerunnable.rerunnable_type == "job":
                ctx.log(f"Job {job_id} was successfully rerun.")
            else:
                ctx.log(
                    f"Job {job_id} associated with {rerunnable.rerunnable_type} {rerunnable.rerunnable_id} was successfully rerun."
                )
    if not job_ids:
        ctx.log(f"No jobs matching the specified {rerunnable.rerunnable_type} {rerunnable.rerunnable_id} were found.")
    run_response = GalaxyBaseRunResponse(
        ctx=ctx,
        runnable=rerunnable,
        user_gi=user_gi,
        history_id=rerunnable.rerunnable_id if rerunnable.rerunnable_type == "history_id" else None,
        log=log_contents_str(config),
    )

    run_response.was_successful = rerun_successful
    return run_response


class GalaxyBaseRunResponse(SuccessfulRunResponse):
    def __init__(
        self,
        ctx,
        runnable,
        user_gi,
        history_id,
        log,
        start_datetime=None,
        end_datetime=None,
    ):
        super().__init__(runnable=runnable)
        self._ctx = ctx
        self._user_gi = user_gi
        self._history_id = history_id
        self._log = log

        self._job_info = None

        self._outputs_dict = None
        self._start_datetime = start_datetime
        self._end_datetime = end_datetime

    def to_galaxy_output(self, output):
        """Convert runnable output to a GalaxyOutput object.

        Subclasses for workflow and tool execution override this.
        """
        raise NotImplementedError()

    @property
    def start_datetime(self):
        """Start datetime of run."""
        return self._start_datetime

    @property
    def end_datetime(self):
        """End datetime of run."""
        return self._end_datetime

    def _get_extra_files(self, dataset_details):
        extra_files_url = (
            f"{self._user_gi.url}/histories/{self._history_id}/contents/{dataset_details['id']}/extra_files"
        )
        extra_files = self._user_gi.jobs._get(url=extra_files_url)
        return extra_files

    def _get_metadata(self, history_content_type, content_id):
        if history_content_type == "dataset":
            return self._user_gi.histories.show_dataset(
                self._history_id,
                content_id,
            )
        elif history_content_type == "dataset_collection":
            return self._user_gi.histories.show_dataset_collection(
                self._history_id,
                content_id,
            )
        else:
            raise Exception("Unknown history content type encountered [%s]" % history_content_type)

    def collect_outputs(self, ctx, output_directory):

        outputs_dict = {}
        # TODO: rather than creating a directory just use
        # Galaxy paths if they are available in this
        # configuration.
        output_directory = output_directory or tempfile.mkdtemp()

        def get_dataset(dataset_details, filename=None):
            parent_basename = dataset_details.get("cwl_file_name") or dataset_details.get("name")
            file_ext = dataset_details["file_ext"]
            if file_ext == "directory":
                # TODO: rename output_directory to outputs_directory because we can have output directories
                # and this is confusing...
                the_output_directory = os.path.join(output_directory, parent_basename)
                safe_makedirs(the_output_directory)
                destination = self.download_output_to(ctx, dataset_details, the_output_directory, filename=filename)
            else:
                destination = self.download_output_to(ctx, dataset_details, output_directory, filename=filename)
            if filename is None:
                basename = parent_basename
            else:
                basename = os.path.basename(filename)

            return {"path": destination, "basename": basename}

        ctx.vlog("collecting outputs to directory %s" % output_directory)

        for runnable_output in get_outputs(self._runnable, gi=self._user_gi):
            output_id = runnable_output.get_id()
            if not output_id:
                ctx.vlog("Workflow output identified without an ID (label), skipping")
                continue
            is_cwl = self._runnable.type in [RunnableType.cwl_workflow, RunnableType.cwl_tool]
            output_src = self.output_src(runnable_output)
            if not output_src:
                # Optional workflow output
                ctx.vlog(f"Optional workflow output '{output_id}' not created, skipping")
                outputs_dict[output_id] = None
                continue
            output_dataset_id = output_src["id"]
            galaxy_output = self.to_galaxy_output(runnable_output)
            cwl_output = output_to_cwl_json(
                galaxy_output,
                self._get_metadata,
                get_dataset,
                self._get_extra_files,
                pseduo_location=True,
            )
            output_dict_value = None
            if is_cwl or output_src["src"] == "hda":
                output_dict_value = cwl_output
            else:

                def attach_file_properties(collection, cwl_output):
                    elements = collection["elements"]
                    assert len(elements) == len(cwl_output)
                    for element, cwl_output_element in zip(elements, cwl_output):
                        element["_output_object"] = cwl_output_element
                        if isinstance(cwl_output_element, list):
                            assert "elements" in element["object"]
                            attach_file_properties(element["object"], cwl_output_element)

                output_metadata = self._get_metadata("dataset_collection", output_dataset_id)
                attach_file_properties(output_metadata, cwl_output)
                output_dict_value = output_metadata

            outputs_dict[output_id] = output_dict_value

        self._outputs_dict = outputs_dict
        ctx.vlog("collected outputs [%s]" % self._outputs_dict)

    @property
    def log(self):
        return self._log

    @property
    def job_info(self):
        print(self._job_info)
        if self._job_info is not None:
            return dict(
                stdout=self._job_info.get("stdout"),
                stderr=self._job_info.get("stderr"),
                command_line=self._job_info.get("command_line"),
            )
        return None

    @property
    def invocation_details(self):
        return None

    @property
    def outputs_dict(self):
        return self._outputs_dict

    def download_output_to(self, ctx, dataset_details, output_directory, filename=None):
        if filename is None:
            local_filename = dataset_details.get("cwl_file_name") or dataset_details.get("name")
        else:
            local_filename = filename
        destination = os.path.join(output_directory, local_filename)
        self._history_content_download(
            ctx,
            self._history_id,
            dataset_details["id"],
            to_path=destination,
            filename=filename,
        )
        return destination

    def _history_content_download(self, ctx, history_id, dataset_id, to_path, filename=None):
        user_gi = self._user_gi
        url = f"{user_gi.url}/histories/{history_id}/contents/{dataset_id}/display"

        data = {}
        if filename:
            data["filename"] = filename

        r = user_gi.make_get_request(url, params=data, stream=True, timeout=user_gi.timeout)
        # Do write an output file before anything could fail, so that the downstream
        # data collection doesn't fail. Comment below still applies.
        with open(to_path, "wb") as fp:
            try:
                r.raise_for_status()
            except HTTPError as e:
                # When a job fails abruptly the object store may not contain a dataset,
                # and that results in an internal server error on the Galaxy side.
                # We don't want this to break the rest of the test report.
                # Should probably find a way to propagate this back into the report.
                ctx.vlog(f"Failed to download history content at URL {url}, exception: {e}")
                return

            for chunk in r.iter_content(chunk_size=bioblend.CHUNK_SIZE):
                if chunk:
                    fp.write(chunk)


class GalaxyToolRunResponse(GalaxyBaseRunResponse):
    def __init__(
        self,
        ctx,
        runnable,
        user_gi,
        history_id,
        log,
        job_info,
        api_run_response,
        start_datetime=None,
        end_datetime=None,
    ):
        super().__init__(
            ctx=ctx,
            runnable=runnable,
            user_gi=user_gi,
            history_id=history_id,
            log=log,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )
        self._job_info = job_info
        self.api_run_response = api_run_response

    def is_collection(self, output):
        # TODO: Make this more rigorous - search both output and output
        # collections - throw an exception if not found in either place instead
        # of just assuming all non-datasets are collections.
        return self.output_src(output)["src"] == "hdca"

    def to_galaxy_output(self, runnable_output):
        output_id = runnable_output.get_id()
        return tool_response_to_output(self.api_run_response, self._history_id, output_id)

    def output_src(self, output):
        outputs = self.api_run_response["outputs"]
        output_collections = self.api_run_response["output_collections"]
        output_id = output.get_id()
        output_src = None
        self._ctx.vlog(f"Looking for id [{output_id}] in outputs [{outputs}]")
        for output in outputs:
            if output["output_name"] == output_id:
                output_src = {"src": "hda", "id": output["id"]}
        for output_collection in output_collections:
            if output_collection["output_name"] == output_id:
                output_src = {"src": "hdca", "id": output_collection["id"]}
        return output_src


class GalaxyWorkflowRunResponse(GalaxyBaseRunResponse):
    def __init__(
        self,
        ctx,
        runnable,
        user_gi,
        history_id,
        log,
        workflow_id,
        invocation_id,
        history_state="ok",
        invocation_state="ok",
        error_message=None,
        start_datetime=None,
        end_datetime=None,
    ):
        super().__init__(
            ctx=ctx,
            runnable=runnable,
            user_gi=user_gi,
            history_id=history_id,
            log=log,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )
        self._workflow_id = workflow_id
        self._invocation_id = invocation_id
        self._invocation_details = {}
        self._cached_invocation = None
        self.history_state = history_state
        self.invocation_state = invocation_state
        self.error_message = error_message
        self._invocation_details = self.collect_invocation_details(invocation_id)

    def to_galaxy_output(self, runnable_output):
        output_id = runnable_output.get_id()
        self._ctx.vlog("checking for output in invocation [%s]" % self._invocation)
        return invocation_to_output(self._invocation, self._history_id, output_id)

    def output_src(self, output):
        invocation = self._invocation
        # Use newer workflow outputs API.

        output_name = output.get_id()
        if output_name in invocation["outputs"]:
            return invocation["outputs"][output.get_id()]
        elif output_name in invocation["output_collections"]:
            return invocation["output_collections"][output.get_id()]
        elif output.is_optional():
            return None
        else:
            raise Exception(f"Failed to find output [{output_name}] in invocation outputs [{invocation['outputs']}]")

    def collect_invocation_details(self, invocation_id=None):
        gi = self._user_gi
        invocation_steps = {}
        invocation = self.get_invocation(invocation_id)
        for step in invocation["steps"]:
            step_label_or_index = f"{step['order_index']}. {step['workflow_step_label'] or 'Unnamed step'}"
            workflow_step = gi.invocations.show_invocation_step(self._invocation_id, step["id"])
            workflow_step["subworkflow"] = None
            subworkflow_invocation_id = workflow_step.get("subworkflow_invocation_id")
            if subworkflow_invocation_id:
                workflow_step["subworkflow"] = self.collect_invocation_details(subworkflow_invocation_id)
            workflow_step_job_details = [
                self._user_gi.jobs.show_job(j["id"], full_details=True) for j in workflow_step["jobs"]
            ]
            workflow_step["jobs"] = workflow_step_job_details
            invocation_steps[step_label_or_index] = workflow_step
        invocation_details = {
            "steps": invocation_steps,
            "details": {
                "invocation_id": self._invocation_id,
                "history_id": self._history_id,
                "workflow_id": self._workflow_id,
                "invocation_state": self.invocation_state,
                "history_state": self.history_state,
                "error_message": self.error_message,
            },
        }
        return invocation_details

    @property
    def invocation_details(self):
        return self._invocation_details

    def get_invocation(self, invocation_id):
        return self._user_gi.invocations.show_invocation(invocation_id)

    @property
    def _invocation(self):
        if self._cached_invocation is None:
            self._cached_invocation = self.get_invocation(self._invocation_id)
        return self._cached_invocation

    @property
    def was_successful(self):
        return self.history_state in ["ok", None] and self.invocation_state == "scheduled"


def _tool_id(tool_path):
    tool_source = get_tool_source(tool_path)
    return tool_source.parse_id()


def _history_id(gi, **kwds):
    history_id = kwds.get("history_id", None)
    if history_id is None:
        history_name = kwds.get("history_name", DEFAULT_HISTORY_NAME) or DEFAULT_HISTORY_NAME
        history_id = gi.histories.create_history(history_name)["id"]
    if kwds.get("tags"):
        tags = kwds.get("tags").split(",")
        gi.histories.update_history(history_id, tags=tags)
    return history_id


def get_dict_from_workflow(gi, workflow_id):
    return gi.workflows.export_workflow_dict(workflow_id)


def _wait_for_invocation(ctx, gi, history_id, workflow_id, invocation_id, polling_backoff=0):
    def state_func():
        return _retry_on_timeouts(ctx, gi, lambda gi: gi.workflows.show_invocation(workflow_id, invocation_id))

    return _wait_on_state(state_func, polling_backoff)


def _retry_on_timeouts(ctx, gi, f):
    gi.timeout = 60
    try_count = 5
    try:
        for try_num in range(try_count):
            start_time = time.time()
            try:
                return f(gi)
            except RequestException:
                end_time = time.time()
                if end_time - start_time > 45 and (try_num + 1) < try_count:
                    ctx.vlog("Galaxy seems to have timedout, retrying to fetch status.")
                    continue
                else:
                    raise
    finally:
        gi.timeout = None


def has_jobs_in_states(ctx, gi, history_id, states):
    params = {"history_id": history_id}
    jobs_url = gi.url + "/jobs"
    jobs = gi.jobs._get(url=jobs_url, params=params)
    target_jobs = [j for j in jobs if j["state"] in states]
    return len(target_jobs) > 0


def _wait_for_history(ctx, gi, history_id, polling_backoff=0):
    # Used to wait for active jobs and then wait for history, but now this is called
    # after upload is complete and after the invocation has been done scheduling - so
    # no need to wait for active jobs anymore I think.

    def state_func():
        return _retry_on_timeouts(ctx, gi, lambda gi: gi.histories.show_history(history_id))

    return _wait_on_state(state_func, polling_backoff)


def _wait_for_invocation_jobs(ctx, gi, invocation_id, polling_backoff=0):
    # Wait for invocation jobs to finish. Less brittle than waiting for a history to finish,
    # as you could have more than one invocation in a history, or an invocation without
    # steps that produce history items.

    ctx.log(f"waiting for invocation {invocation_id}")

    def state_func():
        return _retry_on_timeouts(ctx, gi, lambda gi: gi.jobs.get_jobs(invocation_id=invocation_id))

    return _wait_on_state(state_func, polling_backoff)


def _wait_for_job(gi, job_id, timeout=None):
    def state_func():
        return gi.jobs.show_job(job_id, full_details=True)

    return _wait_on_state(state_func, timeout=timeout)


def _wait_on_state(state_func, polling_backoff=0, timeout=None):
    def get_state():
        response = state_func()
        if not isinstance(response, list):
            response = [response]
        if not response:
            # invocation may not have any attached jobs, that's fine
            return "ok"
        non_terminal_states = {"running", "queued", "new", "ready", "resubmitted", "upload", "waiting"}
        current_states = set(item["state"] for item in response)
        current_non_terminal_states = non_terminal_states.intersection(current_states)
        # Mix of "error"-ish terminal job, dataset, invocation terminal states, so we can use this for whatever we throw at it
        hierarchical_fail_states = [
            "error",
            "paused",
            "deleted",
            "stopped",
            "discarded",
            "failed_metadata",
            "cancelled",
            "failed",
        ]
        for terminal_state in hierarchical_fail_states:
            if terminal_state in current_states:
                # If we got here something has failed and we can return (early)
                return terminal_state
        if current_non_terminal_states:
            return None
        assert len(current_states) == 1, f"unexpected state(s) found: {current_states}"
        return current_states.pop()

    timeout = timeout or 60 * 60 * 24
    final_state = wait_on(get_state, "state", timeout, polling_backoff)
    return final_state


__all__ = ("execute",)



"""Module contains code for the ToolInput class, dealing with the description of tool in workflow and XML."""

from planemo import templates
from planemo.io import info

INPUT_PARAM = """
>{{space}}- *"{{param_label}}"*: `{{param_value}}`
"""

INPUT_FILE_TEMPLATE = """
>{{space}}- {{ '{%' }} icon {{icon}} {{ '%}' }} *"{{input_name}}"*: {{input_value}}
"""

INPUT_SECTION = """
>{{space}}- In *"{{section_label}}"*:
"""

INPUT_ADD_REPEAT = """
>{{space}}- {{ '{%' }} icon param-repeat {{ '%}' }} *"Insert {{repeat_label}}"*
"""

SPACE = "    "


class ToolInput:
    """Class to describe a tool input / parameter and its value from a workflow."""

    def __init__(self, tool_inp_desc, wf_param_values, wf_steps, level, should_be_there=False, force_default=False):
        """Init an instance of ToolInput."""
        self.name = tool_inp_desc["name"]
        if "type" not in tool_inp_desc:
            raise ValueError(f"No type for the parameter {tool_inp_desc['name']}")
        self.type = tool_inp_desc["type"]
        self.tool_inp_desc = tool_inp_desc
        self.level = level
        self.wf_param_values = wf_param_values
        self.wf_steps = wf_steps
        self.formatted_desc = ""
        self.force_default = force_default

        if self.name not in self.wf_param_values:
            if not should_be_there:
                info(f"{self.name} not in workflow")
            else:
                raise ValueError(f"{self.name} not in workflow")
        else:
            self.wf_param_values = self.wf_param_values[self.name]

    def get_formatted_inputs(self):
        """Format the inputs of a step."""
        inputlist = ""
        inps = []
        if isinstance(self.wf_param_values, list):
            # multiple input (not collection)
            icon = "param-files"
            for i in self.wf_param_values:
                inps.append(f"`{i['output_name']}` {get_input_tool_name(i['id'], self.wf_steps)}")
        else:
            inp = self.wf_param_values
            if "id" in inp:
                # sinle input or collection
                inp_type = self.wf_steps[str(inp["id"])]["type"]
                if "collection" in inp_type:
                    icon = "param-collection"
                else:
                    icon = "param-file"
                inps = [f"`{inp['output_name']}` {get_input_tool_name(inp['id'], self.wf_steps)}"]
        if len(inps) > 0:
            inputlist += templates.render(
                INPUT_FILE_TEMPLATE,
                **{
                    "icon": icon,
                    "input_name": self.tool_inp_desc["label"],
                    "input_value": ", ".join(inps),
                    "space": SPACE * self.level,
                },
            )
        return inputlist

    def get_lower_param_desc(self):
        """Get the formatted description of the paramaters in the 'inputs' of the tool description."""
        sub_param_desc = ""
        for inp in self.tool_inp_desc["inputs"]:
            tool_inp = ToolInput(inp, self.wf_param_values, self.wf_steps, self.level + 1)
            sub_param_desc += tool_inp.get_formatted_desc()
        return sub_param_desc

    def get_formatted_section_desc(self):
        """Format the description (label and value) for parameters in a section."""
        section_paramlist = ""
        sub_param_desc = self.get_lower_param_desc()
        if sub_param_desc != "":
            section_paramlist += templates.render(
                INPUT_SECTION, **{"space": SPACE * self.level, "section_label": self.tool_inp_desc["title"]}
            )
            section_paramlist += sub_param_desc
        return section_paramlist

    def get_formatted_conditional_desc(self):
        """Format the description (label and value) for parameters in a conditional."""
        conditional_paramlist = ""
        # Get conditional parameter
        inp = ToolInput(
            self.tool_inp_desc["test_param"],
            self.wf_param_values,
            self.wf_steps,
            self.level,
            should_be_there=True,
            force_default=True,
        )
        conditional_paramlist = inp.get_formatted_desc()
        cond_param = inp.wf_param_values

        # Get parameters in the when and their values
        tmp_tool_inp_desc = self.tool_inp_desc
        for case in tmp_tool_inp_desc["cases"]:
            if case["value"] == cond_param and len(case["inputs"]) > 0:
                self.tool_inp_desc = case
                conditional_paramlist += self.get_lower_param_desc()
        self.tool_inp_desc = tmp_tool_inp_desc
        return conditional_paramlist

    def get_formatted_repeat_desc(self):
        """Format the description (label and value) for parameters in a repeat."""
        repeat_paramlist = ""
        if self.wf_param_values != "[]":
            tool_inp = {}
            for inp in self.tool_inp_desc["inputs"]:
                tool_inp.setdefault(inp["name"], inp)
            tmp_wf_param_values = self.wf_param_values
            cur_level = self.level
            for param in tmp_wf_param_values:
                self.wf_param_values = param
                self.level = cur_level + 1
                paramlist_in_repeat = self.get_lower_param_desc()
                if paramlist_in_repeat != "":
                    # add first click
                    repeat_paramlist += templates.render(
                        INPUT_ADD_REPEAT, **{"space": SPACE * (self.level), "repeat_label": self.tool_inp_desc["title"]}
                    )
                    repeat_paramlist += paramlist_in_repeat
                self.level = cur_level
            self.wf_param_values = tmp_wf_param_values

        repeat_desc = ""
        if repeat_paramlist != "":
            repeat_desc += (
                templates.render(
                    INPUT_SECTION, **{"space": SPACE * self.level, "section_label": self.tool_inp_desc["title"]}
                )
                + repeat_paramlist
            )
        return repeat_desc

    def get_formatted_other_param_desc(self):
        """Get value of a 'simple' parameter if different from the default value, None otherwise."""
        param_value = None
        if self.tool_inp_desc["value"] == self.wf_param_values and not self.force_default:
            param_value = None
        elif self.type == "boolean":
            if bool(self.tool_inp_desc["value"]) == self.wf_param_values:
                param_value = None
            else:
                param_value = "Yes" if self.wf_param_values else "No"
        elif self.type == "select":
            param_values = []
            for opt in self.tool_inp_desc["options"]:
                if opt[1] == self.wf_param_values:
                    param_values.append(opt[0])
            param_value = ", ".join(param_values)
        elif self.type == "data_column":
            param_value = f"c{self.wf_param_values}"
        else:
            param_value = self.wf_param_values

        param_desc = ""
        if param_value is not None:
            param_desc = templates.render(
                INPUT_PARAM,
                **{"space": SPACE * self.level, "param_label": self.tool_inp_desc["label"], "param_value": param_value},
            )
        return param_desc

    def get_formatted_desc(self):
        """Get the formatted description (ready for hands-on tutorial) of the parameter."""
        if self.wf_param_values:
            if self.type == "data" or self.type == "data_collection":
                self.formatted_desc += self.get_formatted_inputs()
            elif self.type == "section":
                self.formatted_desc += self.get_formatted_section_desc()
            elif self.type == "conditional":
                self.formatted_desc += self.get_formatted_conditional_desc()
            elif self.type == "repeat":
                self.formatted_desc += self.get_formatted_repeat_desc()
            else:
                self.formatted_desc += self.get_formatted_other_param_desc()
        return self.formatted_desc


def get_input_tool_name(step_id, steps):
    """Get the string with the name of the tool that generated an input."""
    inp_provenance = ""
    inp_prov_id = str(step_id)
    if inp_prov_id in steps:
        name = steps[inp_prov_id]["name"]
        if "Input dataset" in name:
            inp_provenance = f"({name})"
        else:
            inp_provenance = f"(output of **{name}** {{% icon tool %}})"
    return inp_provenance


def get_empty_input():
    """Get the string for an empty input."""
    return templates.render(
        INPUT_FILE_TEMPLATE,
        **{"space": 1 * SPACE, "icon": "param-file", "input_name": "Input file", "input_value": "File"},
    )


def get_empty_param():
    """Get the string for an empty param."""
    return templates.render(INPUT_PARAM, **{"space": 1 * SPACE, "param_label": "Parameter", "param_value": "a value"})



"""Module describing abstractions for validating XML content."""
import abc
import subprocess
from collections import namedtuple

from galaxy.util import unicodify
from galaxy.util.commands import which

try:
    from lxml import etree
except ImportError:
    etree = None

XMLLINT_COMMAND = "xmllint --noout --schema {0} {1} 2>&1"
INSTALL_VALIDATOR_MESSAGE = (
    "This feature requires an external dependency "
    "to function, pleaes install xmllint (e.g 'brew "
    "install libxml2' or 'apt-get install "
    "libxml2-utils'."
)


class XsdValidator(metaclass=abc.ABCMeta):
    """Class allowing validation of XML files against XSD schema."""

    @abc.abstractmethod
    def validate(self, schema_path, target_path):
        """Validate ``target_path`` against ``schema_path``.

        :return type: ValidationResult
        """

    @abc.abstractmethod
    def enabled(self):
        """Return True iff system has dependencies for this validator.

        :return type: bool
        """


ValidationResult = namedtuple("ValidationResult", ["passed", "output"])


class LxmlValidator(XsdValidator):
    """Validate XSD files using lxml library."""

    def validate(self, schema_path, target_path):
        try:
            xsd_doc = etree.parse(schema_path)
            xsd = etree.XMLSchema(xsd_doc)
            xml = etree.parse(target_path)
            passed = xsd.validate(xml)
            return ValidationResult(passed, xsd.error_log)
        except etree.XMLSyntaxError as e:
            return ValidationResult(False, unicodify(e))

    def enabled(self):
        return etree is not None


class XmllintValidator(XsdValidator):
    """Validate XSD files with the external tool xmllint."""

    def validate(self, schema_path, target_path):
        command = XMLLINT_COMMAND.format(schema_path, target_path)
        p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
        stdout, _ = p.communicate()
        passed = p.returncode == 0
        return ValidationResult(passed, stdout)

    def enabled(self):
        return bool(which("xmllint"))


VALIDATORS = [LxmlValidator(), XmllintValidator()]


def get_validator(require=True):
    """Return a :class:`XsdValidator` object based on available dependencies."""
    for validator in VALIDATORS:
        if validator.enabled():
            return validator

    if require:
        raise Exception(INSTALL_VALIDATOR_MESSAGE)

    return None


__all__ = (
    "get_validator",
    "XsdValidator",
)



"""Module describing the planemo ``list_alias`` command."""
import json

import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import profiles
from planemo.io import info

try:
    from tabulate import tabulate
except ImportError:
    tabulate = None  # type: ignore


@click.command("list_alias")
@options.profile_option(required=True)
@command_function
def cli(ctx, profile, **kwds):
    """
    List aliases for a path or a workflow or dataset ID. Aliases are associated with a particular planemo profile.
    """
    info("Looking for profiles...")
    aliases = profiles.list_alias(ctx, profile)
    if tabulate:
        print(tabulate({"Alias": aliases.keys(), "Object": aliases.values()}, headers="keys"))
    else:
        print(json.dumps(aliases, indent=4, sort_keys=True))

    info(f"{len(aliases)} aliases were found for profile {profile}.")

    ctx.exit(0)
    return



"""Module defines abstractions for configuring Planemo."""

import os
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)

import click
import yaml
from click.core import Option

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext

PLANEMO_CONFIG_ENV_PROP = "PLANEMO_GLOBAL_CONFIG_PATH"
DEFAULT_CONFIG: Dict[str, Any] = {}

VALUE_UNSET = object()


OptionSource = Enum("OptionSource", "cli profile global_config default")


def _default_callback(
    default: Any,
    use_global_config: bool = False,
    resolve_path: bool = False,
    extra_global_config_vars: List[str] = [],
) -> Callable:
    def callback(ctx, param, value):
        planemo_ctx = ctx.obj
        param_name = param.name
        if value is not None:
            result = value
            option_source = OptionSource.cli
        else:
            result, option_source = _find_default(
                planemo_ctx,
                param,
                use_global_config=use_global_config,
                extra_global_config_vars=extra_global_config_vars,
            )

        if result is VALUE_UNSET:
            result = default
            option_source = OptionSource.default

        assert option_source is not None
        assert result is not VALUE_UNSET

        if resolve_path and result is not None:
            result = os.path.abspath(result)

        planemo_ctx.set_option_source(param_name, option_source)
        return result

    return callback


def _find_default(
    ctx: "PlanemoCliContext", param: Option, use_global_config: bool, extra_global_config_vars: List[str]
) -> Union[Tuple[Any, OptionSource], Tuple[object, None]]:
    if use_global_config:
        global_config = ctx.global_config
        global_config_keys = [f"default_{param.name}"] + extra_global_config_vars
        for global_config_key in global_config_keys:
            if global_config_key in global_config:
                default_value = global_config[global_config_key]
                return default_value, OptionSource.global_config

    return VALUE_UNSET, None


def planemo_option(*args, **kwargs) -> Callable:
    """Extend ``click.option`` with planemo-config aware configuration.

    This extends click.option to use a callback when assigning default
    values, add ``use_global_config`` keyword argument to allow reading
    defaults from ~/.planemo.yml, and tracks how parameters are specified
    using the Planemo Context object.
    """
    option_type = kwargs.get("type")
    use_global_config = kwargs.pop("use_global_config", False)
    use_env_var = kwargs.pop("use_env_var", False)
    extra_global_config_vars = kwargs.pop("extra_global_config_vars", [])

    default_specified = "default" in kwargs
    default = None
    if default_specified:
        default = kwargs.pop("default")

    if default_specified or use_global_config or use_env_var:
        outer_callback = kwargs.pop("callback", None)

        def callback(ctx, param, value):
            resolve_path = bool(option_type and getattr(option_type, "resolve_path", False))
            result = _default_callback(
                default,
                use_global_config=use_global_config,
                extra_global_config_vars=extra_global_config_vars,
                resolve_path=resolve_path,
            )(ctx, param, value)

            if outer_callback is not None:
                result = outer_callback(ctx, param, result)

            return result

        kwargs["callback"] = callback

    if default_specified:
        kwargs["default"] = None

    if use_env_var:
        name = None
        for arg in args:
            if arg.startswith("--"):
                name = arg[len("--") :]
        assert name
        kwargs["envvar"] = f"PLANEMO_{name.upper()}"

    option = click.option(*args, **kwargs)
    return option


def global_config_path(config_path: Optional[str] = None) -> str:
    if not config_path:
        planemo_config_env_prop = os.environ.get(PLANEMO_CONFIG_ENV_PROP, "~/.planemo.yml")
        config_path = os.path.expanduser(planemo_config_env_prop)
    return config_path


def read_global_config(config_path: Optional[str]) -> Dict[str, Any]:
    config_path = global_config_path(config_path)
    if not os.path.exists(config_path):
        return DEFAULT_CONFIG

    with open(config_path) as f:
        return yaml.safe_load(f)


__all__ = (
    "global_config_path",
    "read_global_config",
    "planemo_option",
)



"""Module describing the planemo ``shed_init`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function


@click.command("shed_init")
@options.optional_project_arg()
@click.option(
    "--from_workflow",
    type=click.Path(exists=True, file_okay=True, resolve_path=True),
    help=("Attempt to generate repository dependencies from specified " "workflow."),
)
@click.option("--description", help="Specify repository description for .shed.yml.")
@click.option("--long_description", help="Specify repository long_description for .shed.yml.")
@click.option("--remote_repository_url", help="Specify repository remote_repository_url for .shed.yml.")
@click.option("--homepage_url", help="Specify repository homepage_url for .shed.yml.")
@click.option(
    "--category",
    multiple=True,
    help="Specify repository category for .shed.yml (may specify multiple).",
    type=click.Choice(shed.CURRENT_CATEGORIES),
)
@options.shed_repo_options()
@options.force_option()
@command_function
def cli(ctx, path, **kwds):
    """Bootstrap new Tool Shed .shed.yml file.

    This Tool Shed configuration file is used by other ``planemo`` commands
    such as ``shed_lint``, ``shed_create``, ``shed_upload``, and ``shed_diff``
    to manage repositories in a Galaxy Tool Shed.
    """
    exit_code = shed.shed_init(ctx, path, **kwds)
    sys.exit(exit_code)



"""Module describing the planemo ``normalize`` command."""
from xml.etree import ElementTree

import click
from galaxy.tool_util.linters.xml_order import TAG_ORDER
from galaxy.tool_util.loader import (
    load_tool,
    raw_tool_xml_tree,
)
from galaxy.util import Element

from planemo import options
from planemo.cli import command_function


@click.command("normalize")
@options.required_tool_arg()
@click.option(
    "--expand_macros",
    is_flag=True,
    help="Expand macros while normalizing tool XML - useful to see how macros are evaluated.",
)
@click.option(
    "--skip_reorder",
    is_flag=True,
    help=(
        "Planemo will reorder top-level tool blocks according to tool "
        "development best practices as part of this command, this flag "
        "will disable that behavior."
    ),
)
@click.option(
    "--skip_reindent",
    is_flag=True,
    help=(
        "Planemo will reindent the XML according to tool development "
        "best practices as part of this command, this flag will disable "
        "that behavior."
    ),
)
@command_function
def cli(ctx, path, expand_macros=False, **kwds):
    """Generate normalized tool XML from input.

    This will break the formatting of your tool and is currently only intended
    for viewing macro expansions for for use with XSD validation (see
    https://github.com/JeanFred/Galaxy-XSD for instance). Please do not use
    the output as is - it frequently makes tool less readable not more.

    The top-level blocks will be reordered and whitespace fixed according to
    the tool development best practices outlined on the Galaxy wiki.

    \b
        % # Print normalized version of tool.
        % planemo normalize tool.xml
        <tool>
        ...
        % # Print a variant of tool with all macros expanded out, useful for
        % # debugging complex macros.
        % planemo normalize --expand_macros tool.xml
        <tool>
        ...
    """
    if expand_macros:
        tree = load_tool(path)
    else:
        tree = raw_tool_xml_tree(path)

    root = tree.getroot()
    if not kwds.get("skip_reorder", False):
        last_index = len(TAG_ORDER)
        data = []
        for elem in root:
            tag = elem.tag
            if tag in TAG_ORDER:
                key = TAG_ORDER.index(tag)
            else:
                key = last_index
                last_index += 1
            data.append((key, elem))
        data.sort()
        root[:] = [item[-1] for item in data]
    if not kwds.get("skip_reindent", False):
        _indent(root)
    ElementTree.dump(root)


def _indent(elem: Element, level: int = 0) -> None:
    # http://stackoverflow.com/questions/749796/pretty-printing-xml-in-python
    i = "\n" + level * "    "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "    "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for subelem in elem:
            _indent(subelem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i



import socket
from http.client import BadStatusLine
from time import time as now
from urllib.error import URLError
from urllib.request import urlopen


def get_free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("localhost", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def wait_http_service(url, timeout=None):
    if timeout:
        end = now() + timeout

    while True:
        try:
            if timeout:
                next_timeout = end - now()
                if next_timeout < 0:
                    return False

            kwds = {} if timeout is None else dict(timeout=next_timeout)
            with urlopen(url, **kwds) as r:
                if r.getcode() != 200:
                    continue
            return True
        except OSError:
            pass
        except BadStatusLine:
            pass
        except URLError:
            pass


# code.activestate.com/recipes/576655-wait-for-network-service-to-appear
def wait_net_service(server, port, timeout=None):
    """Wait for network service to appear.

    :param int timeout: in seconds, if None or 0 wait forever
    :return: A ``bool`` - if ``timeout`` is ``None`` may return only ``True`` or
             throw an unhandled network exception.
    """
    if port is None:
        raise TypeError("wait_net_service passed NoneType port value.")

    port = int(port)

    if timeout:
        end = now() + timeout

    while True:
        s = socket.socket()
        # Following line prevents this method from interfering with process
        # it is waiting for on localhost.
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            if timeout:
                next_timeout = end - now()
                if next_timeout < 0:
                    return False
                else:
                    s.settimeout(next_timeout)

            s.connect((server, port))

        except socket.timeout:
            # this exception occurs only if timeout is set
            if timeout:
                return False

        except OSError:
            # if getattr(e, "errno") == 61:
            #    refused_connections += 1
            s.close()
        else:
            s.close()
            return True



"""Utilities for interacting with Github."""

import io
import os
import stat
import tarfile
import tempfile
import time
from distutils.dir_util import copy_tree
from pathlib import Path

import requests

from planemo import git
from planemo.io import (
    communicate,
    IS_OS_X,
)

try:
    import github

    has_github_lib = True
except ImportError:
    github = None
    has_github_lib = False

GH_VERSION = "1.5.0"

NO_GITHUB_DEP_ERROR = "Cannot use github functionality - " "PyGithub library not available."
FAILED_TO_DOWNLOAD_GH = "No gh executable available and it could not be installed."
DEFAULT_REMOTE_NAME = "planemo-remote"
SLEEP_BEFORE_RELEASE = int(os.environ.get("PLANEMO_SLEEP_BEFORE_RELEASE", 60))


def get_github_config(ctx, allow_anonymous=False):
    """Return a :class:`planemo.github_util.GithubConfig` for given configuration."""
    global_github_config = _get_raw_github_config(ctx)
    return GithubConfig(global_github_config, allow_anonymous=allow_anonymous)


def clone_fork_branch(ctx, target, path, remote_name=DEFAULT_REMOTE_NAME, **kwds):
    """Clone, fork, and branch a repository ahead of building a pull request."""
    git.checkout(ctx, target, path, branch=kwds.get("branch", None), remote="origin", from_branch="master")
    if kwds.get("fork"):
        try:
            fork(ctx, path, remote_name=remote_name, **kwds)
        except Exception:
            pass
    if "GITHUB_USER" in os.environ:
        # On CI systems fork doesn't add a local remote under circumstances I don't quite understand,
        # but that's probably linked to https://github.com/cli/cli/issues/2722
        cmd = [
            "git",
            "remote",
            "add",
            remote_name,
            f"https://github.com/{os.environ['GITHUB_USER']}/{os.path.basename(target)}",
        ]
        try:
            communicate(cmd, cwd=path)
        except RuntimeError:
            # Can add the remote only once
            pass
    return remote_name


def fork(ctx, path, remote_name=DEFAULT_REMOTE_NAME, **kwds):
    """Fork the target repository using ``gh``."""
    gh_path = ensure_gh(ctx, **kwds)
    gh_env = get_gh_env(ctx, path, **kwds)
    cmd = [gh_path, "repo", "fork", "--remote=true", "--remote-name", remote_name]
    communicate(cmd, cwd=path, env=gh_env)
    return remote_name


def get_or_create_repository(ctx, owner, repo, dry_run=True, **kwds):
    """Clones or creates a repository and returns path on disk"""
    target = os.path.realpath(tempfile.mkdtemp())
    remote_repo = f"https://github.com/{owner}/{repo}"
    try:
        ctx.log(f"Cloning {remote_repo}")
        git.clone(ctx, src=remote_repo, dest=target)
    except Exception:
        ctx.log(f"Creating repository {remote_repo}")
        target = create_repository(ctx, owner=owner, repo=repo, dest=target, dry_run=dry_run)
    return target


def create_repository(ctx, owner, repo, dest, dry_run, **kwds):
    gh_path = ensure_gh(ctx, **kwds)
    gh_env = get_gh_env(ctx, dry_run=dry_run, **kwds)
    cmd = [gh_path, "repo", "create", "-y", "--public", f"{owner}/{repo}"]
    if dry_run:
        "Would run command '{}'".format(" ".join(cmd))
        git.init(ctx, dest)
        return dest
    communicate(cmd, env=gh_env, cwd=dest)
    return os.path.join(dest, repo)


def rm_dir_contents(directory, ignore_dirs=(".git")):
    directory = Path(directory)
    for item in directory.iterdir():
        if item.name not in ignore_dirs:
            if item.is_dir():
                rm_dir_contents(item)
            else:
                item.unlink()


def add_dir_contents_to_repo(
    ctx, from_dir, target_dir, target_repository_path, version, dry_run, branch="main", notes=""
):
    ctx.log(f"From {from_dir} to {target_repository_path}")
    rm_dir_contents(target_repository_path)
    copy_tree(from_dir, target_repository_path)
    git.add(ctx, target_repository_path, target_repository_path)
    message = f"Update for version {version}"
    if notes:
        message += f"\n{notes}"
    git.commit(ctx, repo_path=target_repository_path, message=message)
    if not dry_run:
        git.push(ctx, target_repository_path, to="origin", branch=branch)


def assert_new_version(ctx, version, owner, repo):
    remote_repo = f"https://github.com/{owner}/{repo}"
    try:
        tags_and_versions = git.ls_remote(ctx, remote_repo=remote_repo)
        if f"refs/tags/v{version}" in tags_and_versions or f"refs/tags/{version}" in tags_and_versions:
            raise Exception(f"Version '{version}' for {owner}/{repo} exists already. Please change the version.")
    except RuntimeError:
        # repo doesn't exist
        pass


def changelog_in_repo(target_repository_path):
    """
    Parse the changelog for the current version.

    The changelog should be written following the conventions in https://keepachangelog.com/en/0.3.0/.
    This means we will assume that individual versions are announced using `##`.
    """
    changelog = []
    for path in os.listdir(target_repository_path):
        if "changelog.md" in path.lower():
            version_header_seen = False
            version_header_prefix = "## "
            top_level_header_seen = False
            top_level_header_prefix = "# "
            with open(os.path.join(target_repository_path, path)) as changelog_fh:
                for line in changelog_fh:
                    if line.startswith(top_level_header_prefix):
                        top_level_header_seen = True
                        continue
                    if line.startswith(version_header_prefix):
                        if version_header_seen:
                            return "\n".join(changelog[:-1])
                        else:
                            version_header_seen = True
                    if top_level_header_seen and not line or not version_header_seen:
                        continue
                    changelog.append(line.rstrip())
    return "\n".join(changelog).rstrip()


def create_release(ctx, from_dir, target_dir, owner, repo, version, dry_run, branch, notes="", **kwds):
    assert_new_version(ctx, version, owner=owner, repo=repo)
    target_repository_path = get_or_create_repository(ctx, owner=owner, repo=repo, dry_run=dry_run)
    add_dir_contents_to_repo(
        ctx, from_dir, target_dir, target_repository_path, version=version, dry_run=dry_run, branch=branch, notes=notes
    )
    gh_path = ensure_gh(ctx, **kwds)
    gh_env = get_gh_env(ctx, dry_run=dry_run, **kwds)
    cmd = [
        gh_path,
        "release",
        "-R",
        f"{owner}/{repo}",
        "create",
        f"v{version}",
        "--title",
        str(version),
    ]
    cmd.extend(["--notes", notes or changelog_in_repo(target_repository_path)])
    if not dry_run:
        # For new repositories dockstore needs a bit of time to register the new workflow.
        time.sleep(SLEEP_BEFORE_RELEASE)
        communicate(cmd, env=gh_env)
    else:
        ctx.log("Would run command '{}'".format(" ".join(cmd)))


def pull_request(ctx, path, message=None, repo=None, **kwds):
    """Create a pull request against the origin of the path using ``gh``."""
    gh_path = ensure_gh(ctx, **kwds)
    gh_env = get_gh_env(ctx, path, **kwds)
    cmd = [gh_path, "pr", "create"]
    if message is None:
        cmd.append("--fill")
    else:
        lines = message.splitlines()
        cmd.extend(["--title", lines[0]])
        if len(lines) > 1:
            cmd.extend(["--body", "\n".join(lines[1:])])
    if repo:
        cmd.extend(["--repo", repo])
    communicate(cmd, env=gh_env)


def get_gh_env(ctx, path=None, dry_run=False, **kwds):
    """Return a environment dictionary to run gh with given user and repository target."""
    if path is None:
        env = {}
    else:
        env = git.git_env_for(path).copy()
    if not dry_run:
        github_config = _get_raw_github_config(ctx)
        if github_config is not None:
            if "access_token" in github_config:
                env["GITHUB_TOKEN"] = github_config["access_token"]

    return env


def ensure_gh(ctx, **kwds):
    """Ensure gh is available for planemo

    This method will ensure ``gh`` is installed at the correct version.

    For more information on ``gh`` checkout https://cli.github.com/
    """
    planemo_gh_path = os.path.join(ctx.workspace, f"gh-{GH_VERSION}")
    if not os.path.exists(planemo_gh_path):
        _try_download_gh(planemo_gh_path)

    if not os.path.exists(planemo_gh_path):
        raise Exception(FAILED_TO_DOWNLOAD_GH)

    return planemo_gh_path


def _try_download_gh(planemo_gh_path):
    link = _gh_link()
    path = Path(planemo_gh_path)
    resp = requests.get(link)
    with tarfile.open(fileobj=io.BytesIO(resp.content)) as tf, path.open("wb") as outfile:
        for member in tf.getmembers():
            if member.name.endswith("bin/gh"):
                outfile.write(tf.extractfile(member).read())
    path.chmod(path.stat().st_mode | stat.S_IEXEC)


def _get_raw_github_config(ctx):
    """Return a :class:`planemo.github_util.GithubConfig` for given configuration."""
    if "github" not in ctx.global_config:
        if "GITHUB_TOKEN" in os.environ:
            return {
                "access_token": os.environ["GITHUB_TOKEN"],
            }
    if "github" not in ctx.global_config:
        raise Exception("github account not found in planemo config and GITHUB_TOKEN environment variables unset")
    return ctx.global_config["github"]


class GithubConfig:
    """Abstraction around a Github account.

    Required to use ``github`` module methods that require authorization.
    """

    def __init__(self, config, allow_anonymous=False):
        if not has_github_lib:
            raise Exception(NO_GITHUB_DEP_ERROR)
        if "access_token" not in config:
            if not allow_anonymous:
                raise Exception("github authentication unavailable")
            github_object = github.Github()
        else:
            github_object = github.Github(config["access_token"])
        self._github = github_object


def _gh_link():
    if IS_OS_X:
        template_link = "https://github.com/cli/cli/releases/download/v%s/gh_%s_macOS_amd64.tar.gz"
    else:
        template_link = "https://github.com/cli/cli/releases/download/v%s/gh_%s_linux_amd64.tar.gz"
    return template_link % (GH_VERSION, GH_VERSION)


def publish_as_gist_file(ctx, path, name="index"):
    """Publish a gist.

    More information on gists at http://gist.github.com/.
    """
    github_config = get_github_config(ctx, allow_anonymous=False)
    user = github_config._github.get_user()
    with open(path) as fh:
        content = fh.read()
    content_file = github.InputFileContent(content)
    gist = user.create_gist(False, {name: content_file})
    return gist.files[name].raw_url


def get_repository_object(ctx, name):
    github_object = get_github_config(ctx, allow_anonymous=True)
    return github_object._github.get_repo(name)


__all__ = (
    "add_dir_contents_to_repo",
    "clone_fork_branch",
    "create_release",
    "ensure_gh",
    "fork",
    "get_github_config",
    "get_gh_env",
    "get_or_create_repository",
    "publish_as_gist_file",
)



"""Templating abstraction around jinja2 for Planemo."""
try:
    from jinja2 import Template
except ImportError:
    Template = None  # type: ignore

NO_JINJA2_MESSAGE = (
    "This functionality requires Jinja2 but this library is " "unavailable. Install with `pip install jinja2`."
)


def render(template_str: str, **kwds):
    """Use jinja2 to render specified template."""
    if Template is None:
        raise Exception(NO_JINJA2_MESSAGE)
    template = Template(template_str)
    contents = template.render(**kwds)
    return contents



""" Tool linting module that lints Galaxy tools for their DOIs (if a DOI type citation is present)
"""
import planemo.lint


def lint_tool_dois(tool_xml, lint_ctx):
    planemo.lint.lint_dois(tool_xml, lint_ctx)



"""Module describing the planemo ``test`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.engine.test import test_runnables
from planemo.io import temp_directory
from planemo.runnable import RunnableType
from planemo.runnable_resolve import for_runnable_identifiers


@click.command("test")
@options.optional_tools_arg(multiple=True, allow_uris=True)
@click.option(
    "--failed",
    is_flag=True,
    help="Re-run only failed tests. This command will read "
    "tool_test_output.json to determine which tests failed so this "
    "file must have been produced with the same set of tool ids "
    "previously.",
    default=False,
)
@click.option(
    "--polling_backoff",
    type=int,
    help="Poll resources with an increasing interval between requests. "
    "Useful when testing against remote and/or production "
    "instances to limit generated traffic.",
    default="0",
)
@options.galaxy_target_options()
@options.galaxy_config_options()
@options.test_options()
@options.engine_options()
@command_function
def cli(ctx, uris, **kwds):
    """Run specified tool's tests within Galaxy.

    All referenced tools (by default all the tools in the current working
    directory) will be tested and the results quickly summarized.

    To run these tests planemo needs a Galaxy instance to utilize, planemo
    will search parent directories to see if any is a Galaxy instance
    - but one can pick the Galaxy instance to use with the --galaxy_root
    option or force planemo to download a disposable instance with the
    ``--install_galaxy`` flag.

    In addition to to quick summary printed to the console - various detailed
    output summaries can be configured. ``tool_test_output.html`` (settable
    via ``--test_output``) will contain a human consumable HTML report
    describing the test run. A JSON file (settable via ``--test_output_json``
    and defaulting to ``tool_test_output.json``) will also be created. These
    files can can be disabled by passing in empty arguments or globally by
    setting the values ``default_test_output`` and/or
    ``default_test_output_json`` in ``~/.planemo.yml`` to ``null``. For
    continuous integration testing a xUnit-style report can be configured using
    the ``--test_output_xunit``.

    planemo uses temporarily generated config files and environment variables
    to attempt to shield this execution of Galaxy from manually launched runs
    against that same Galaxy root - but this may not be bullet proof yet so
    please careful and do not try this against production Galaxy instances.
    """
    with temp_directory(dir=ctx.planemo_directory) as temp_path:
        # Create temp dir(s) outside of temp, docker can't mount $TEMPDIR on OSX
        runnables = for_runnable_identifiers(ctx, uris, kwds, temp_path=temp_path)

        # pick a default engine type if needed
        is_cwl = all(r.type in {RunnableType.cwl_tool, RunnableType.cwl_workflow} for r in runnables)
        if kwds.get("engine", None) is None:
            if is_cwl:
                kwds["engine"] = "cwltool"
            elif kwds.get("galaxy_url", None):
                kwds["engine"] = "external_galaxy"
            else:
                kwds["engine"] = "galaxy"

        return_value = test_runnables(ctx, runnables, original_paths=uris, **kwds)

    ctx.exit(return_value)



"""Module describing the planemo ``syntax`` command."""
import click

from planemo.cli import command_function

SYNTAX_URL = "https://docs.galaxyproject.org/en/latest/dev/schema.html"


@click.command("syntax")
@command_function
def cli(ctx, **kwds):
    """Open tool config syntax page in web browser."""
    click.launch(SYNTAX_URL)



"""Module describing the planemo ``shed_diff`` command."""
import shutil
import sys
import tempfile
from xml.sax.saxutils import escape

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function
from planemo.io import captured_io_for_xunit
from planemo.reports.xunit_handler import handle_report_xunit_kwd


@click.command("shed_diff")
@options.shed_read_options()
@click.option(
    "-o",
    "--output",
    type=click.Path(file_okay=True, resolve_path=True),
    help="Send diff output to specified file.",
    default=None,
)
@click.option(
    "--shed_target_source",
    help="Source Tool Shed to diff against (will ignore local project info"
    " specified). To compare the main Tool Shed against the test, set"
    " this to testtoolshed.",
    default=None,
)
@click.option(
    "--raw",
    is_flag=True,
    help="Do not attempt smart diff of XML to filter out attributes " "populated by the Tool Shed.",
)
@options.report_xunit()
@command_function
def cli(ctx, paths, **kwds):
    """diff between local repository and Tool Shed.

    By default, this will produce a diff between this repository and what
    would be uploaded to the Tool Shed with the `shed_upload` command - but
    this command can be made to compare other combinations of repositories.
    Here are some examples

    \b
        $ # diff for this repository and the main Tool Shed
        $ planemo shed_diff
        $ # diff for this repository and the test Tool Shed
        $ planemo shed_diff --shed_target testtoolshed
        $ # diff for the test Tool Shed and main Tool Shed
        $ planemo shed_diff --shed_target_source testtoolshed
        $ # diff for two an explicitly specified repositories (ignores
        $ # current project's shed YAML file.)
        $ planemo shed_diff --owner peterjc --name blast_rbh
            --shed_target_source testtoolshed

    This command will return an exit code of:

    - 0 if there are no detected differences.
    - 1 if there are differences.
    - 2 if the target repository doesn't exist.
    - >200 if there are errors attempting to perform a diff.

    **Warning:** ``shed_diff`` doesn't inspect repository metadata, this
    difference applies only to the file contents of files that would actually be
    uploaded to the repository.
    """

    # In a little bit of cheating, we're defining this variable here to collect
    # a "report" on our shed_diff
    collected_data = {
        "results": {
            "total": 0,
            "errors": 0,
            "failures": 0,
            "skips": 0,
        },
        "suitename": "shed_diff",
        "tests": [],
    }

    def diff(realized_repository):

        # We create a temporary redirection from kwds's
        # output to our tempfile. This lets us capture the
        # diff and redirect it to their requested location as
        # well as to the XUnit report.
        diff_output = tempfile.NamedTemporaryFile(mode="r")
        user_requested_output = kwds.get("output", None)
        # Replace their output handle with ours
        kwds["output"] = diff_output.name

        captured_io = {}
        with captured_io_for_xunit(kwds, captured_io):
            result = shed.diff_repo(ctx, realized_repository, **kwds)

        # May be extraneous but just want to ensure entire file is written
        # before a copy is made.
        diff_output.flush()
        # Redirect a copy to user_requested_output if they did:
        if user_requested_output is not None:
            shutil.copy(diff_output.name, user_requested_output)
        else:
            with open(diff_output.name) as f:
                sys.stdout.write(f.read())
        # Rewind to the start of the file and read it in its entirety
        diff_output.seek(0)
        diff_output_contents = diff_output.read()
        diff_output.close()

        # Collect data about what happened
        collected_data["results"]["total"] += 1
        xunit_case = {
            "name": "shed-diff",
            "classname": realized_repository.name,
            "time": captured_io["time"],
            "stdout": captured_io["stdout"],
            "stderr": captured_io["stderr"],
        }
        if result >= 200:
            collected_data["results"]["errors"] += 1
            xunit_case.update(
                {
                    "errorType": "DiffError",
                    "errorMessage": "Error diffing repositories",
                    "errorContent": escape(diff_output_contents),
                    "time": captured_io["time"],
                }
            )
        elif result > 2:
            collected_data["results"]["failures"] += 1
            xunit_case.update(
                {
                    "errorType": "PlanemoDiffError",
                    "errorMessage": "Planemo error diffing repositories",
                    "errorContent": escape(diff_output_contents),
                }
            )
        elif result == 2:
            collected_data["results"]["failures"] += 1
            xunit_case.update(
                {
                    "errorType": "RepoDoesNotExist",
                    "errorMessage": "Target Repository does not exist",
                    "errorContent": escape(diff_output_contents),
                }
            )
        elif result == 1:
            collected_data["results"]["failures"] += 1
            xunit_case.update(
                {
                    "errorType": "Different",
                    "errorMessage": "Repository is different",
                    "errorContent": escape(diff_output_contents),
                }
            )

        # Append our xunit test case
        collected_data["tests"].append(xunit_case)

        return result

    exit_code = shed.for_each_repository(ctx, diff, paths, **kwds)

    handle_report_xunit_kwd(kwds, collected_data)

    sys.exit(exit_code)



import base64

from galaxy.util import strip_control_characters
from jinja2 import (
    Environment,
    PackageLoader,
)
from pkg_resources import resource_string

TITLE = "Results (powered by Planemo)"


def build_report(structured_data, report_type="html", execution_type="Test", **kwds):
    """Use report_{report_type}.tpl to build page for report."""
    environment = dict(
        title=TITLE,
        raw_data=structured_data,
    )

    __fix_test_ids(environment)
    environment = __inject_summary(environment)

    if report_type == "html":
        # The HTML report format needs a lot of extra, custom data.
        # IMO, this seems to suggest it should be embedded.
        environment["title"] = None
        environment["execution_type"] = execution_type
        markdown = template_data(environment, "report_markdown.tpl")
        environment["title"] = " ".join((environment["execution_type"], TITLE))
        environment["raw_data"] = base64.b64encode(markdown.encode("utf-8")).decode("utf-8")
        environment.update(
            {
                "custom_style": __style("custom.css"),
                "custom_script": __script("custom"),
                "bootstrap_style": __style("bootstrap.min.css"),
                "jquery_script": __script("jquery.min"),
                "bootstrap_script": __script("bootstrap.min"),
                "markdown_it_script": __script("markdown-it.min"),
            }
        )

    return template_data(environment, "report_%s.tpl" % report_type)


def template_data(environment, template_name, **kwds):
    """Build an arbitrary templated page."""
    env_kwargs = {}
    if template_name == "report_markdown.tpl":
        env_kwargs["keep_trailing_newline"] = True
        env_kwargs["trim_blocks"] = True
    env = Environment(loader=PackageLoader("planemo", "reports"), **env_kwargs)
    env.filters["strip_control_characters"] = lambda x: strip_control_characters(x) if x else x
    template = env.get_template(template_name)
    return template.render(**environment)


def __fix_test_ids(environment):
    for test in environment["raw_data"]["tests"]:
        test_data = test.get("data")
        if test_data and test_data.get("tool_id"):
            test["id"] = f"{test_data['tool_id']} (Test #{test_data['test_index'] + 1})"


def __inject_summary(environment):
    total = 0
    errors = 0
    failures = 0
    skips = 0
    for execution in environment["raw_data"]["tests"]:
        total += 1
        test_data = execution.get("data")
        if test_data:
            status = test_data.get("status")
            if status == "error":
                errors += 1
            elif status == "failure":
                failures += 1
            elif status == "skipped":
                skips += 1
    environment["raw_data"]["results"] = {
        "total": total,
        "errors": errors,
        "failures": failures,
        "skips": skips,
    }
    if "suitename" not in environment:
        environment["raw_data"]["suitename"] = TITLE
    return environment


def __style(filename):
    resource = __load_resource(filename)
    return "<style>%s</style>" % resource


def __script(short_name):
    resource = __load_resource("%s.js" % short_name)
    return "<script>%s</script>" % resource


def __load_resource(name):
    return resource_string(__name__, name).decode("UTF-8")



"""Tool linting module that lints Galaxy tool against experimental XSD."""
import copy
import os
import tempfile

import galaxy.tool_util

import planemo.lint

TOOL_XSD = os.path.join(os.path.dirname(galaxy.tool_util.__file__), "xsd", "galaxy.xsd")


def lint_tool_xsd(tool_xml, lint_ctx):
    """Write a temp file out and lint it."""
    with tempfile.NamedTemporaryFile() as tf:
        _clean_root(tool_xml).write(tf.name)
        planemo.lint.lint_xsd(lint_ctx, TOOL_XSD, tf.name)


def _clean_root(tool_xml):
    """XSD assumes macros have been expanded, so remove them."""
    clean_tool_xml = copy.deepcopy(tool_xml)
    to_remove = []
    for macros_el in clean_tool_xml.getroot().findall("macros"):
        to_remove.append(macros_el)
    for macros_el in to_remove:
        clean_tool_xml.getroot().remove(macros_el)
    return clean_tool_xml



"""Ensure best-practice biocontainer registered for this tool."""

from galaxy.tool_util.deps.container_resolvers.mulled import targets_to_mulled_name
from galaxy.tool_util.deps.mulled.util import build_target

from planemo.conda import tool_source_conda_targets

MESSAGE_WARN_NO_REQUIREMENTS = "No valid package requirement tags found to infer BioContainer from."
MESSAGE_WARN_NO_CONTAINER = "Failed to find a BioContainer registered for these requirements."
MESSAGE_INFO_FOUND_BIOCONTAINER = "BioContainer best-practice container found [%s]."

lint_tool_types = ["*"]


def lint_biocontainer_registered(tool_source, lint_ctx):
    conda_targets = tool_source_conda_targets(tool_source)
    if not conda_targets:
        lint_ctx.warn(MESSAGE_WARN_NO_REQUIREMENTS)
        return

    mulled_targets = [build_target(c.package, c.version) for c in conda_targets]
    name = mulled_container_name("biocontainers", mulled_targets)
    if name:
        lint_ctx.info(MESSAGE_INFO_FOUND_BIOCONTAINER % name)
    else:
        lint_ctx.warn(MESSAGE_WARN_NO_CONTAINER)


def mulled_container_name(namespace, targets):
    name = targets_to_mulled_name(targets=targets, hash_func="v2", namespace=namespace)
    if name:
        return f"quay.io/{namespace}/{name}"



"""The module describes a CLI framework extending ``click``."""
import functools
import os
import sys
from typing import (
    Any,
    Callable,
    Dict,
    List,
)

import click
from click.core import (
    Command,
    Context,
)

from planemo import __version__
from planemo.context import (
    configure_standard_planemo_logging,
    PlanemoContext,
)
from planemo.exit_codes import ExitCodeException
from planemo.galaxy import profiles
from .config import OptionSource

CONTEXT_SETTINGS = dict(auto_envvar_prefix="PLANEMO")
COMMAND_ALIASES = {
    "l": "lint",
    "o": "open",
    "t": "test",
    "s": "serve",
}


class PlanemoCliContext(PlanemoContext):
    """Describe context of Planemo CLI computation.

    Extend PlanemoContext with operations for CLI concerns (exit code) and
    for interacting with the click library.
    """

    def _log_message(self, message: str) -> None:
        click.echo(message, file=sys.stderr)

    def exit(self, exit_code: int):
        """Exit planemo with the supplied exit code."""
        self.vlog("Exiting planemo with exit code [%d]" % exit_code)
        raise ExitCodeException(exit_code)


pass_context = click.make_pass_decorator(PlanemoCliContext, ensure=True)
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))


def list_cmds() -> List[str]:
    """List planemo commands from commands folder."""
    rv = []
    for filename in os.listdir(cmd_folder):
        if filename.endswith(".py") and filename.startswith("cmd_"):
            rv.append(filename[len("cmd_") : -len(".py")])
    rv.sort()
    return rv


def name_to_command(name: str) -> Command:
    """Convert a subcommand name to the cli function for that command.

    Command <X> is defined by the method 'planemo.commands.cmd_<x>:cli',
    this method uses `__import__` to load and return that method.
    """
    try:
        mod_name = "planemo.commands.cmd_" + name
        mod = __import__(mod_name, None, None, ["cli"])
    except ImportError as e:
        raise Exception(f"Problem loading command {name}, exception {e}")
    return mod.cli


class PlanemoCLI(click.MultiCommand):
    def list_commands(self, ctx: Context) -> List[str]:
        return list_cmds()

    def get_command(self, ctx: Context, name: str) -> Command:
        if name in COMMAND_ALIASES:
            name = COMMAND_ALIASES[name]
        return name_to_command(name)


def command_function(f: Callable) -> Callable:
    """Extension point for processing kwds after click callbacks."""

    @functools.wraps(f)
    def handle_blended_options(*args, **kwds):
        profile = kwds.get("profile", None)
        if profile:
            ctx = args[0]
            profile_defaults = profiles.ensure_profile(ctx, profile, **kwds)
            _setup_profile_options(ctx, profile_defaults, kwds)

        try:
            return f(*args, **kwds)
        except ExitCodeException as e:
            sys.exit(e.exit_code)

    return pass_context(handle_blended_options)


def _setup_profile_options(ctx: PlanemoCliContext, profile_defaults: Dict[str, Any], kwds: Dict[str, Any]) -> None:
    for key, value in profile_defaults.items():
        option_present = key in kwds
        option_cli_specified = option_present and (ctx.get_option_source(key) == OptionSource.cli)
        use_profile_option = not option_present or not option_cli_specified
        if use_profile_option:
            kwds[key] = value
            ctx.set_option_source(key, OptionSource.profile, force=True)


@click.command(cls=PlanemoCLI, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__)
@click.option("-v", "--verbose", is_flag=True, help="Enables verbose mode.")
@click.option(
    "--config", default="~/.planemo.yml", envvar="PLANEMO_GLOBAL_CONFIG_PATH", help="Planemo configuration YAML file."
)
@click.option("--directory", default="~/.planemo", envvar="PLANEMO_GLOBAL_WORKSPACE", help="Workspace for planemo.")
@pass_context
def planemo(ctx, config, directory, verbose, configure_logging=True):
    """A command-line toolkit for building tools and workflows for Galaxy.

    Check out the full documentation for Planemo online
    http://planemo.readthedocs.org or open with ``planemo docs``.

    All the individual planemo commands support the ``--help`` option, for
    example use ``planemo lint --help`` for more details on checking tools.
    """
    ctx.verbose = verbose
    if configure_logging:
        configure_standard_planemo_logging(verbose)
    ctx.planemo_config = os.path.expanduser(config)
    ctx.planemo_directory = os.path.expanduser(directory)


__all__ = (
    "command_function",
    "list_cmds",
    "name_to_command",
    "planemo",
    "PlanemoCliContext",
)



"""Module contianing the :class:`ToilEngine` implementation of :class:`Engine`."""

from planemo import cwl
from planemo.runnable import RunnableType
from .interface import BaseEngine


class ToilEngine(BaseEngine):
    """An :class:`Engine` implementation backed by Toil.

    More information on toil can be found at https://github.com/BD2KGenomics/toil.
    """

    handled_runnable_types = [RunnableType.cwl_tool, RunnableType.cwl_workflow]

    def _run(self, runnables, job_paths):
        """Run CWL job using Toil."""
        results = []
        for runnable, job_path in zip(runnables, job_paths):
            results.append(cwl.run_toil(self._ctx, runnable, job_path, **self._kwds))
        return results


__all__ = ("ToilEngine",)



from planemo.engine import engine_context
from planemo.galaxy.config import _find_test_data
from planemo.galaxy.test import handle_reports_and_summary
from planemo.runnable import for_paths


def test_runnables(ctx, runnables, original_paths=None, **kwds):
    """Return exit code indicating test or failure."""
    if kwds.get("update_test_data"):
        non_copied_runnables = for_paths(original_paths)
        kwds["test_data_target_dir"] = _find_test_data(non_copied_runnables, **kwds)
    with engine_context(ctx, **kwds) as engine:
        test_data = engine.test(runnables, test_timeout=kwds.get("test_timeout"))
        ctx.vlog(f"engine.test returning [{test_data}]")
        return handle_reports_and_summary(ctx, test_data.structured_data, kwds=kwds)



"""Interface over bioblend and direct access to ToolShed API via requests."""

import json

from galaxy.util import unicodify

from planemo.bioblend import (
    ensure_module,
    toolshed,
)
from planemo.io import untar_to

REPOSITORY_DOWNLOAD_TEMPLATE = "%srepository/download?repository_id=%s" "&changeset_revision=default&file_type=gz"


def tool_shed_instance(url, key, email, password):
    ensure_module()
    tsi = toolshed.ToolShedInstance(url=url, key=key, email=email, password=password)
    return tsi


def find_repository(tsi, owner, name):
    """Find repository information for given owner and repository
    name.
    """
    repos = tsi.repositories.get_repositories()

    def matches(r):
        return r["owner"] == owner and r["name"] == name

    matching_repos = list(filter(matches, repos))
    if not matching_repos:
        return None
    else:
        return matching_repos[0]


def latest_installable_revision(tsi, repository_id):
    info = tsi.repositories.show_repository(repository_id)
    owner = info["owner"]
    name = info["name"]
    revisions = tsi.repositories.get_ordered_installable_revisions(name, owner)
    if len(revisions) == 0:
        msg = f"Failed to find installable revisions for [{owner}, {name}]."
        raise Exception(msg)
    else:
        return revisions[-1]


def username(tsi):
    """Fetch current username from shed given API key/auth."""
    user = _user(tsi)
    return user["username"]


def api_exception_to_message(e):
    """Convert API exception to human digestable error message - parsing
    out the shed generate message if possible.
    """
    message = unicodify(e)
    if hasattr(e, "read"):
        message = e.read()
        try:
            # Galaxy passes nice JSON messages as their errors, which bioblend
            # blindly returns. Attempt to parse those.
            upstream_error = json.loads(message)
            message = upstream_error["err_msg"]
        except Exception:
            pass
    return message


def find_category_ids(tsi, categories):
    """Translate human readable category names into their associated IDs."""
    category_list = tsi.categories.get_categories()

    category_ids = []
    for cat in categories:
        matching_cats = [x for x in category_list if x["name"] == cat]
        if not matching_cats:
            message = "Failed to find category %s" % cat
            raise Exception(message)
        category_ids.append(matching_cats[0]["id"])
    return category_ids


def download_tar(tsi, repo_id, destination, to_directory):
    base_url = tsi.base_url
    if not base_url.endswith("/"):
        base_url += "/"
    download_url = REPOSITORY_DOWNLOAD_TEMPLATE % (base_url, repo_id)
    if to_directory:
        tar_args = ["-xzf", "-", "--strip-components=1"]
        untar_to(download_url, tar_args=tar_args, dest_dir=destination)
    else:
        untar_to(download_url, path=destination)


def _user(tsi):
    """Fetch user information from the ToolShed API for given
    key.
    """
    # TODO: this should be done with an actual bioblend method,
    # see https://github.com/galaxyproject/bioblend/issues/130.
    response = tsi.make_get_request(tsi.url + "/users")
    return response.json()[0]



"""Module describing the planemo ``clone`` command."""
from typing import Callable

import click

from planemo import (
    github_util,
    options,
)
from planemo.cli import command_function
from planemo.config import planemo_option

CLONE_GITHUB_TARGETS = {
    "tools-iuc": "galaxyproject/tools-iuc",
    "tools-devteam": "galaxyproject/tools-devteam",
    "galaxy": "galaxyproject/galaxy",
    "planemo": "galaxyproject/planemo",
    "tools-galaxyp": "galaxyproteomics/tools-galaxyp",
    "bioconda-recipes": "bioconda/bioconda-recipes",
    "homebrew-science": "Homebrew/homebrew-science",
    "workflows": "common-workflow-language/workflows",
}


def clone_target_arg() -> Callable:
    """Represent target to clone/branch."""
    return click.argument(
        "target",
        metavar="TARGET",
        type=click.STRING,
    )


@click.command("clone")
@planemo_option(
    "--fork/--skip_fork",
    default=True,
    is_flag=True,
)
@planemo_option("--branch", type=click.STRING, default=None, help="Create a named branch on result.")
@clone_target_arg()
@options.optional_project_arg(exists=None, default="__NONE__")
@command_function
def cli(ctx, target, path, **kwds):
    """Short-cut to quickly clone, fork, and branch a relevant Github repo.

    For instance, the following will clone, fork, and branch the tools-iuc
    repository to allow a subsequent pull request to fix a problem with bwa.


    \b
        $ planemo clone --branch bwa-fix tools-iuc
        $ cd tools-iuc
        $ # Make changes.
        $ git add -p # Add desired changes.
        $ git commit -m "Fix bwa problem."
        $ planemo pull_request -m "Fix bwa problem."

    These changes do require that a github access token is
    specified in ~/.planemo.yml. An access token can be generated by going
    to https://github.com/settings/tokens.
    """
    if target in CLONE_GITHUB_TARGETS:
        target = "https://github.com/%s" % CLONE_GITHUB_TARGETS[target]
    # Pretty hacky that this path isn't treated as None.
    if path is None or path.endswith("__NONE__"):
        path = target.split("/")[-1]
    github_util.clone_fork_branch(ctx, target, path, **kwds)



"""Module describing the planemo ``container_register`` command."""
import os
import string

import click
from galaxy.tool_util.deps.container_resolvers.mulled import targets_to_mulled_name
from galaxy.tool_util.deps.mulled.mulled_build import (
    base_image_for_targets,
    DEFAULT_BASE_IMAGE,
)
from galaxy.tool_util.deps.mulled.util import (
    conda_build_target_str,
    v2_image_name,
)

from planemo import options
from planemo.cli import command_function
from planemo.conda import (
    best_practice_search,
    build_conda_context,
    collect_conda_target_lists_and_tool_paths,
)
from planemo.git import (
    add,
    branch,
    commit,
    push,
)
from planemo.github_util import (
    clone_fork_branch,
    DEFAULT_REMOTE_NAME,
    get_repository_object,
    pull_request,
)
from planemo.mulled import conda_to_mulled_targets

REGISTRY_TARGET_NAME = "multi-package-containers"
REGISTRY_TARGET_PATH = "combinations"
REGISTRY_REPOSITORY = "BioContainers/multi-package-containers"
DEFAULT_MESSAGE = "Add container $hash.\n**Hash**: $hash\n\n**Packages**:\n$packages\nBase Image:$base_image\n\n"
DEFAULT_MESSAGE += "**For** :\n$tools\n\nGenerated with Planemo."
CONTENTS = "#targets\tbase_image\timage_build\n$targets\t$base_image\t$image_build\n"
BIOCONTAINERS_PLATFORM = "linux-64"


@click.command("container_register")
@options.optional_tools_arg(multiple=True)
@options.recursive_option()
@options.mulled_namespace_option()
@options.conda_target_options()
@click.option(
    "output_directory",
    "--output_directory",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        resolve_path=True,
    ),
    default=None,
    help=("Container registration directory (defaults to ~/.planemo/multi-package-containers."),
)
@click.option(
    "-m",
    "--message",
    default=DEFAULT_MESSAGE,
    help="Commit and pull request message template for registration interactions.",
)
@click.option(
    "--pull_request/--no_pull_request",
    is_flag=True,
    default=True,
    help="Fork and create a pull request against %s for these changes." % REGISTRY_REPOSITORY,
)
@click.option(
    "--force_push/--no_force_push",
    is_flag=True,
    default=False,
    help="Force push branch for pull request in case it already exists.",
)
@command_function
def cli(ctx, paths, **kwds):
    """Register multi-requirement containers as needed.

    BioContainers publishes all Bioconda packages automatically as individual
    container images. These however are not enough for tools with multiple
    best-practice requirements. Such requirements should be recorded and published
    so that a container can be created and registered for these tools.
    """
    registry_target = RegistryTarget(ctx, **kwds)
    conda_context = build_conda_context(ctx, **kwds)

    combinations_added = 0
    conda_targets_list, tool_paths_list = collect_conda_target_lists_and_tool_paths(
        ctx, paths, recursive=kwds["recursive"]
    )
    for conda_targets, tool_paths in zip(conda_targets_list, tool_paths_list):
        ctx.vlog("Handling conda_targets [%s]" % conda_targets)
        mulled_targets = conda_to_mulled_targets(conda_targets)
        mulled_targets_str = "- " + "\n- ".join(map(conda_build_target_str, mulled_targets))

        if len(mulled_targets) < 1:
            ctx.log("Skipping registration, no targets discovered.")
            continue

        name = v2_image_name(mulled_targets)
        tag = "0"
        name_and_tag = f"{name}-{tag}"
        target_filename = os.path.join(registry_target.output_directory, "%s.tsv" % name_and_tag)
        if os.path.exists(target_filename):
            ctx.log("Target file '%s' already exists, skipping" % target_filename)
            continue

        if targets_to_mulled_name(mulled_targets, hash_func="v2", namespace=kwds["mulled_namespace"]):
            ctx.vlog("quay repository already exists, skipping")
            continue

        if registry_target.has_pull_request_for(name):
            ctx.vlog("Found matching open pull request for [%s], skipping" % name)
            continue

        best_practice_requirements = True
        for conda_target in conda_targets:
            best_hit, exact = best_practice_search(
                conda_target, conda_context=conda_context, platform=BIOCONTAINERS_PLATFORM
            )
            if not best_hit:
                ctx.log("Target [%s] is not available in best practice channels - skipping" % conda_target)
                best_practice_requirements = False
            if not exact:
                ctx.log(
                    "Target version [%s] for package [%s] is not available in best practice channels - please specify the full version",
                    conda_target.version,
                    conda_target.package,
                )

        if not best_practice_requirements:
            continue

        base_image = DEFAULT_BASE_IMAGE
        for conda_target in conda_targets:
            base_image = base_image_for_targets([conda_target], conda_context=conda_context)
            if base_image != DEFAULT_BASE_IMAGE:
                ctx.log(f"{conda_target} requires '{base_image}' as base image")
                break

        registry_target.write_targets(ctx, target_filename, mulled_targets, tag, base_image)
        tools_str = "\n".join(map(lambda p: "- " + os.path.basename(p), tool_paths))
        registry_target.handle_pull_request(
            ctx, name, target_filename, mulled_targets_str, tools_str, base_image, **kwds
        )
        combinations_added += 1


class RegistryTarget:
    """Abstraction around mulled container registry (both directory and Github repo)."""

    def __init__(self, ctx, **kwds):
        output_directory = kwds["output_directory"]
        pr_titles = []
        target_repository = None
        self.remote_name = DEFAULT_REMOTE_NAME
        do_pull_request = kwds.get("pull_request", True)
        if output_directory is None:
            target_repository = os.path.join(ctx.workspace, REGISTRY_TARGET_NAME)
            output_directory = os.path.join(target_repository, REGISTRY_TARGET_PATH)
            self.remote_name = (
                clone_fork_branch(
                    ctx,
                    "https://github.com/%s" % REGISTRY_REPOSITORY,
                    target_repository,
                    fork=do_pull_request,
                )
                or self.remote_name
            )
            pr_titles = [pr.title for pr in open_prs(ctx)]

        self.do_pull_request = do_pull_request
        self.pr_titles = pr_titles
        self.output_directory = output_directory
        self.target_repository = target_repository

    def has_pull_request_for(self, name):
        has_pr = False
        if self.do_pull_request:
            if any([name in t for t in self.pr_titles]):
                has_pr = True

        return has_pr

    def handle_pull_request(self, ctx, name, target_filename, packages_str, tools_str, base_image, **kwds):
        if self.do_pull_request:
            message = kwds["message"]
            message = string.Template(message).safe_substitute(
                {
                    "hash": name,
                    "packages": packages_str,
                    "tools": tools_str,
                    "base_image": base_image,
                }
            )
            branch_name = name.replace(":", "-")
            branch(ctx, self.target_repository, branch_name, from_branch="master")
            add(ctx, self.target_repository, target_filename)
            commit(ctx, self.target_repository, message=message)
            force_push = kwds.get("force_push", False)
            push(ctx, repo_path=self.target_repository, to=self.remote_name, branch=branch_name, force=force_push)
            pull_request(ctx, self.target_repository, message=message, repo=REGISTRY_REPOSITORY)

    def write_targets(self, ctx, target_filename, mulled_targets, tag, base_image):
        with open(target_filename, "w") as f:
            targets = to_target_str(mulled_targets)
            f.write(string.Template(CONTENTS).safe_substitute(targets=targets, base_image=base_image, image_build=tag))
            ctx.log(f"Wrote requirements [{targets}] to file [{target_filename}]")


def to_target_str(targets):
    target_strings = []
    for target in targets:
        if target.version:
            target_str = f"{target.package_name}={target.version}"
        else:
            target_str = target.package_name
        target_strings.append(target_str)
    return ",".join(target_strings)


def open_prs(ctx):
    repo = get_repository_object(ctx, REGISTRY_REPOSITORY)
    prs = [pr for pr in repo.get_pulls()]
    return prs



"""Module describing the planemo ``mull`` command."""
import click
from galaxy.tool_util.deps.mulled.mulled_build import mull_targets

from planemo import options
from planemo.cli import command_function
from planemo.mulled import (
    build_mull_target_kwds,
    collect_mulled_target_lists,
)


@click.command("mull")
@options.optional_tools_arg(multiple=True)
@options.recursive_option()
@options.mulled_options()
@options.conda_ensure_channels_option()
@command_function
def cli(ctx, paths, **kwds):
    """Build containers for specified tools.

    Supplied tools will be inspected for referenced requirement packages. For
    each combination of requirements a "mulled" container will be built. Galaxy
    can automatically discover this container and subsequently use it to run
    or test the tool.

    For this to work, the tool's requirements will need to be present in a known
    Conda channel such as bioconda (https://github.com/bioconda/bioconda-recipes).
    This can be verified by running ``planemo lint --conda_requirements`` on the
    target tool(s).
    """
    for mulled_targets in collect_mulled_target_lists(ctx, paths, recursive=kwds["recursive"]):
        mull_target_kwds = build_mull_target_kwds(ctx, **kwds)
        command = kwds["mulled_command"]
        mull_targets(mulled_targets, command=command, **mull_target_kwds)



"""Module describing the planemo ``docker_shell`` command.

The docker_shell command takes the path to a single tool file and returns
a new command to launch a new shell into the corresponding Docker container
for that tool.

Galaxy will treat the identifier in the tool XML file as a published Docker
image - during development such the image an image may not yet exist or may
be not updated. So during development the --from_tag argument to used to treat
the identifier as a locally cached tag. (Tip: Use the docker_build to populate
such a tag from a Dockerfile located in the tool's directory.
"""

import os

import click
from galaxy.tool_util.deps import (
    docker_util,
    dockerfiles,
)
from galaxy.tool_util.deps.requirements import parse_requirements_from_xml
from galaxy.tool_util.loader import load_tool

from planemo import options
from planemo.cli import command_function


@click.command("docker_shell")
@options.required_tool_arg()
@click.option(
    "--from_tag", is_flag=True, help=("Treat the tool's Docker container identifier as a locally cached tag.")
)
@click.option("--shell", default="/bin/bash", help="Shell to launch in container (defaults to /bin/bash).")
@options.docker_cmd_option()
@options.docker_sudo_option()
@options.docker_sudo_cmd_option()
@options.docker_host_option()
@command_function
def cli(ctx, path, **kwds):
    """Launch shell in Docker container for a tool.

    Will launch a shell in the Docker container referenced by the specified
    tool. Prints a command to do this the way Galaxy would in job files it
    generates - so be sure to wrap this in $(...) to launch the subshell.

    \b
        $ $(planemo docker_shell bowtie2.xml)
        ...
        root@b8754062f875:/#

    """
    tool_xml = load_tool(path)
    requirements, containers = parse_requirements_from_xml(tool_xml)
    identifier = None
    for container in containers:
        if container.type == "docker":
            identifier = container.identifier

    # Refactor mulled container resolver to be able to do this.
    if kwds["from_tag"]:
        identifier = "-t %s" % identifier

    tool_dir = os.path.dirname(os.path.abspath(path))
    working_dir = os.path.abspath(os.getcwd())
    if tool_dir.startswith(working_dir):
        volumes = [f"{working_dir}:{working_dir}"]
    elif working_dir.startswith(tool_dir):
        volumes = [f"{tool_dir}:{tool_dir}"]
    else:
        volumes = [f"{working_dir}:{working_dir}", f"{tool_dir}:{tool_dir}"]

    script = docker_util.build_docker_run_command(
        "/bin/bash",
        identifier,
        interactive=True,
        terminal=True,
        working_directory=working_dir,
        volumes=volumes,
        **dockerfiles.docker_host_args(**kwds),
    )
    print(script)



"""Actions related to running and reporting on Galaxy-specific testing."""

import json

import click
from galaxy.util import unicodify

from planemo.exit_codes import (
    EXIT_CODE_GENERIC_FAILURE,
    EXIT_CODE_NO_SUCH_TARGET,
    EXIT_CODE_OK,
)
from planemo.io import (
    info,
    warn,
)
from planemo.reports import (
    allure,
    build_report,
)
from planemo.test.results import get_dict_value
from . import structures as test_structures

NO_XUNIT_REPORT_MESSAGE = (
    "Cannot locate xUnit report [%s] for tests - " "required to build planemo report and summarize " "tests."
)
NO_JSON_REPORT_MESSAGE = (
    "Cannot locate JSON report [%s] for tests - " "required to build planemo report and summarize " "tests."
)
REPORT_NOT_CHANGED = (
    "Galaxy failed to update test report [%s] for tests - " "required to build planemo report and summarize " "tests."
)
NO_TESTS_MESSAGE = "No tests were executed - see Galaxy output for details."
ALL_TESTS_PASSED_MESSAGE = "All %d test(s) executed passed."
PROBLEM_COUNT_MESSAGE = (
    "There were problems with %d test(s) - out of %d " "test(s) executed. See %s for detailed breakdown."
)
GENERIC_PROBLEMS_MESSAGE = "One or more tests failed. See %s for detailed " "breakdown."
GENERIC_TESTS_PASSED_MESSAGE = "No failing tests encountered."
TEST_DATA_UPDATED_MESSAGE = "Test data were updated and tests were rerun."
TEST_DATA_NOT_UPDATED_MESSAGE = "%s Therefore, no test data were updated." % ALL_TESTS_PASSED_MESSAGE


def handle_reports_and_summary(ctx, structured_data, exit_code=None, kwds=None):
    """Produce reports and print summary, return 0 if tests passed.

    If ``exit_code`` is set - use underlying test source for return
    code and test success determination, otherwise infer from supplied
    test data.
    """
    if kwds is None:
        kwds = {}
    handle_reports(ctx, structured_data, kwds)
    summary_exit_code = _handle_summary(structured_data, **kwds)
    return exit_code if exit_code is not None else summary_exit_code


def merge_reports(input_paths, output_path):
    reports = []
    for path in input_paths:
        with open(path, encoding="utf-8") as f:
            reports.append(json.load(f))
    tests = []
    for report in reports:
        tests.extend(report["tests"])
    tests = sorted(tests, key=lambda k: k["id"])
    merged_report = {"tests": tests}
    with open(output_path, mode="w", encoding="utf-8") as out:
        out.write(unicodify(json.dumps(merged_report)))


def handle_reports(ctx, structured_data, kwds):
    """Write reports based on user specified kwds."""
    exceptions = []
    structured_report_file = kwds.get("test_output_json", None)
    if structured_report_file:
        try:
            with open(structured_report_file, mode="w", encoding="utf-8") as f:
                f.write(unicodify(json.dumps(structured_data)))
        except Exception as e:
            exceptions.append(e)

    for report_type in ["html", "markdown", "text", "xunit", "junit", "allure"]:
        try:
            _handle_test_output_file(ctx, report_type, structured_data, kwds)
        except Exception as e:
            exceptions.append(e)

    if len(exceptions) > 0:
        raise exceptions[0]


def _handle_test_output_file(ctx, report_type, test_data, kwds):
    kwd_name = "test_output"
    if report_type != "html":
        kwd_name = "test_output_%s" % report_type

    path = kwds.get(kwd_name, None)
    if path is None:
        message = "No file specified for %s, skipping test output." % kwd_name
        ctx.vlog(message)
        return

    if report_type == "allure":
        file_modication_datatime = kwds.get("file_modication_datatime")
        allure.write_results(path, test_data, file_modication_datatime=file_modication_datatime)
        return

    execution_type = kwds.get("execution_type", "Test")
    try:
        contents = build_report.build_report(test_data, report_type=report_type, execution_type=execution_type)
    except Exception:
        message = f"Problem producing report file {path} for {kwd_name}"
        ctx.vlog(message, exception=True)
        raise

    try:
        with open(path, mode="w", encoding="utf-8") as handle:
            handle.write(unicodify(contents))
    except Exception:
        message = f"Problem writing output file {kwd_name} for {path}"
        ctx.vlog(message, exception=True)
        raise


def _handle_summary(structured_data, **kwds):
    summary_dict = get_dict_value("summary", structured_data)
    num_tests = get_dict_value("num_tests", summary_dict)
    num_failures = get_dict_value("num_failures", summary_dict)
    num_errors = get_dict_value("num_errors", summary_dict)
    num_problems = num_failures + num_errors

    summary_exit_code = EXIT_CODE_OK
    if num_problems > 0:
        summary_exit_code = EXIT_CODE_GENERIC_FAILURE
    elif num_tests == 0:
        summary_exit_code = EXIT_CODE_NO_SUCH_TARGET

    summary_style = kwds.get("summary")
    if kwds.get("test_data_updated"):
        info(TEST_DATA_UPDATED_MESSAGE)
    if summary_style != "none":
        if num_tests == 0:
            warn(NO_TESTS_MESSAGE)
        elif num_problems == 0:
            if kwds.get("update_test_data") and not kwds.get("test_data_updated"):
                info(TEST_DATA_NOT_UPDATED_MESSAGE % num_tests)
            else:
                info(ALL_TESTS_PASSED_MESSAGE % num_tests)
        elif num_problems:
            html_report_file = kwds.get("test_output")
            message_args = (num_problems, num_tests, html_report_file)
            message = PROBLEM_COUNT_MESSAGE % message_args
            warn(message)

        _summarize_tests_full(structured_data, **kwds)

    return summary_exit_code


def _summarize_tests_full(structured_data, **kwds):
    tests = get_dict_value("tests", structured_data)
    for test_case_data in tests:
        _summarize_test_case(test_case_data, **kwds)


def passed(xunit_testcase_el):
    did_pass = True
    for child_el in list(xunit_testcase_el):
        if child_el.tag in ["failure", "error"]:
            did_pass = False
    return did_pass


def _summarize_test_case(structured_data, **kwds):
    summary_style = kwds.get("summary")
    test_id = test_structures.case_id(raw_id=get_dict_value("id", structured_data))
    status = get_dict_value("status", get_dict_value("data", structured_data))
    if status != "success":
        state = click.style("failed", bold=True, fg="red")
    else:
        state = click.style("passed", bold=True, fg="green")
    click.echo(test_id.label + ": " + state)
    if summary_style != "minimal":
        _print_command_line(structured_data, test_id)


def _print_command_line(test, test_id):
    execution_problem = test.get("execution_problem", None)
    if execution_problem:
        click.echo("| command: *could not execute job, no command generated* ")
        return

    job = None
    try:
        job = test["job"]
    except (KeyError, IndexError):
        click.echo("| command: *failed to find job for test object [%s]" % test)
        return
    try:
        command = job["command_line"]
    except (KeyError, IndexError):
        click.echo("| command: *failed to find command_line for job object [%s]" % job)
        return

    click.echo("| command: %s" % command)


__all__ = (
    "handle_reports",
    "handle_reports_and_summary",
)



import json
import tempfile

try:
    from toil.cwl import cwltoil
except ImportError:
    cwltoil = None

from planemo.deps import ensure_dependency_resolvers_conf_configured
from planemo.io import (
    error,
    real_io,
)
from planemo.runnable import (
    ErrorRunResponse,
    Runnable,
)
from .run import (
    CwlToolRunResponse,
    JSON_PARSE_ERROR_MESSAGE,
)

TOIL_REQUIRED_MESSAGE = "This functionality requires Toil, please install with 'pip install toil'"


def run_toil(ctx, runnable: "Runnable", job_path: str, **kwds):
    """Translate planemo kwds to cwltool kwds and run cwltool main function."""
    _ensure_toil_available()

    args = []
    if not ctx.verbose:
        args.append("--quiet")
    output_directory = kwds.get("output_directory", None)
    if output_directory:
        args.append("--outdir")
        args.append(output_directory)
    if kwds.get("no_container", False):
        args.append("--no-container")
        ensure_dependency_resolvers_conf_configured(ctx, kwds)
        args.append("--beta-dependency-resolvers-configuration")
        args.append(kwds["dependency_resolvers_config_file"])
    if kwds.get("mulled_containers"):
        args.append("--beta-use-biocontainers")

    if kwds.get("non_strict_cwl", False):
        args.append("--non-strict")

    args.extend([runnable.path, job_path])
    ctx.vlog(f"Calling cwltoil with arguments {args}")
    with tempfile.NamedTemporaryFile("w") as tmp_stdout:
        # cwltool passes sys.stderr to subprocess.Popen - ensure it has
        # and actual fileno.
        with real_io():
            ret_code = cwltoil.main(args, stdout=tmp_stdout)
        tmp_stdout.flush()
        with open(tmp_stdout.name) as stdout_f:
            try:
                result = json.load(stdout_f)
            except ValueError:
                message = JSON_PARSE_ERROR_MESSAGE % (
                    open(tmp_stdout.name).read(),
                    tmp_stdout.name,
                )
                error(message)
                raise Exception(message)

        if ret_code != 0:
            return ErrorRunResponse("Error running Toil")
        outputs = result
    return CwlToolRunResponse(
        runnable,
        "",
        outputs=outputs,
    )


def _ensure_toil_available():
    if cwltoil is None:
        raise Exception(TOIL_REQUIRED_MESSAGE)



"""Abstractions for serving out development Galaxy servers."""

import contextlib
import os
import time

from planemo import (
    io,
    network_util,
)
from .config import galaxy_config
from .ephemeris_sleep import sleep
from .run import run_galaxy_command


@contextlib.contextmanager
def serve(ctx, runnables=None, **kwds):
    if runnables is None:
        runnables = []
    """Serve a Galaxy instance with artifacts defined by paths."""
    try:
        with _serve(ctx, runnables, **kwds) as config:
            yield config
    except Exception as e:
        ctx.vlog("Problem serving Galaxy", exception=e)
        raise


@contextlib.contextmanager
def _serve(ctx, runnables, **kwds):
    engine = kwds.get("engine", "galaxy")
    if engine == "docker_galaxy":
        kwds["dockerize"] = True

    daemon = kwds.get("daemon", False)
    if daemon:
        kwds["no_cleanup"] = True

    port = kwds.get("port", None)
    if port is None:
        port = network_util.get_free_port()
        kwds["port"] = port

    with galaxy_config(ctx, runnables, **kwds) as config:
        cmd = config.startup_command(ctx, **kwds)
        action = "Starting Galaxy"
        exit_code = run_galaxy_command(
            ctx,
            cmd,
            config.env,
            action,
        )
        if exit_code:
            message = "Problem running Galaxy command [%s]." % config.log_contents
            io.warn(message)
            raise Exception(message)
        host = kwds.get("host", "127.0.0.1")

        startup_timeout = kwds.get("galaxy_startup_timeout", 900)
        galaxy_url = f"http://{host}:{port}"
        galaxy_alive = sleep(galaxy_url, verbose=ctx.verbose, timeout=startup_timeout)
        if not galaxy_alive:
            raise Exception(
                f"Attempted to serve Galaxy at {galaxy_url}, but it failed to start in {startup_timeout} seconds."
            )
        config.install_workflows()
        if kwds.get("pid_file"):
            real_pid_file = config.pid_file
            if os.path.exists(config.pid_file):
                os.symlink(real_pid_file, kwds["pid_file"])
            else:
                io.warn("Can't find Galaxy pid file [%s] to link" % real_pid_file)
        yield config


@contextlib.contextmanager
def serve_daemon(ctx, runnables=None, **kwds):
    """Serve a daemonized Galaxy instance with artifacts defined by paths."""
    if runnables is None:
        runnables = []
    config = None
    kwds["daemon"] = True
    try:
        with serve(ctx, runnables, **kwds) as config:
            yield config
    finally:
        if config:
            if ctx.verbose:
                print("Galaxy Log:")
                print(config.log_contents)
            config.kill()
            if not kwds.get("no_cleanup", False):
                config.cleanup()


def sleep_for_serve():
    # This is bad, do something better...
    for _ in range(3600 * 24):
        time.sleep(1)


__all__ = (
    "serve",
    "serve_daemon",
)



"""Module describing the planemo ``dockstore_init`` command."""
import os

import click

from planemo import options
from planemo.cli import command_function
from planemo.workflow_lint import (
    DOCKSTORE_REGISTRY_CONF,
    generate_dockstore_yaml,
)


@click.command("dockstore_init")
@options.optional_project_arg()
@options.publish_dockstore_option()
@command_function
def cli(ctx, path=".", **kwds):
    """Initialize a .dockstore.yml configuration file for workflows in directory.

    Walk supplied directory and find all Galaxy workflows and test configurations
    and create a ``.dockstore.yml`` with references to these files. Be sure to push
    this file to Github before registering your workflow repository with Dockstore.

    Visit Dockstore at https://dockstore.org/. Find more about registering workflows
    with Dockstore at
    https://docs.dockstore.org/en/develop/getting-started/dockstore-workflows.html.
    """
    # TODO: implement -f semantics
    dockstore_path = os.path.join(path, DOCKSTORE_REGISTRY_CONF)
    contents = generate_dockstore_yaml(path, kwds["publish"])
    with open(dockstore_path, "w") as f:
        f.write(contents)



"""Module describing the planemo ``shed_lint`` command."""
import click

from planemo import (
    options,
    shed,
    shed_lint,
)
from planemo.cli import command_function


@click.command("shed_lint")
@options.shed_realization_options()
@options.report_level_option()
@options.fail_level_option()
@options.click.option(
    "--tools", is_flag=True, default=False, help=("Lint tools discovered in the process of linting repositories.")
)
@options.lint_xsd_option()
@options.click.option(
    "--ensure_metadata",
    is_flag=True,
    default=False,
    help=(
        "Ensure .shed.yml files contain enough metadata for each repository "
        "to allow automated creation and/or updates."
    ),
)
@click.option(
    "--urls",
    is_flag=True,
    default=False,
    help="Check validity of URLs in XML files",
)
# @click.option(
#     "--verify",
#     is_flag=True,
#     help="If an sha256sum is available, download the entire file AND validate it.",
#     default=False,
# )
@command_function
def cli(ctx, paths, **kwds):
    """Check Tool Shed repository for common issues.

    With the ``--tools`` flag, this command lints actual Galaxy tools
    in addition to tool shed artifacts.

    With the ``--urls`` flag, this command searches for
    ``<package>$URL</package>`` and download actions which specify URLs. Each
    of those are accessed individually. By default, this tool requests the
    first hundred or so bytes of each listed URL and validates that a 200 OK
    was received. In tool XML files, the ``--urls`` option checks through the
    help text for mentioned URLs and checks those.
    """

    def lint(realized_repository):
        return shed_lint.lint_repository(ctx, realized_repository, **kwds)

    kwds["fail_on_missing"] = False
    exit_code = shed.for_each_repository(ctx, lint, paths, **kwds)
    ctx.exit(exit_code)



import json
import re
import uuid

from allure_commons import plugin_manager
from allure_commons.lifecycle import AllureLifecycle
from allure_commons.logger import AllureFileLogger
from allure_commons.model2 import (
    Label,
    Link,
    Status,
    StatusDetails,
)
from allure_commons.types import (
    AttachmentType,
    LabelType,
    LinkType,
)
from allure_commons.utils import (
    escape_non_unicode_symbols,
    md5,
    platform_label,
    uuid4,
)
from dateutil import parser
from galaxy.util import safe_makedirs

JSON_INDENT = 2
WORKFLOW_INDEX_MATCH = re.compile(r"(.*)\_([\d+])")


class AllureListener:
    def __init__(self, lifecycle):
        self.lifecycle = lifecycle


class AllureWriter:
    def __init__(self, results_path):
        safe_makedirs(results_path)
        self.lifecycle = AllureLifecycle()
        self.logger = AllureFileLogger(results_path)
        self.listener = AllureListener(self.lifecycle)

    def process(self, structured_data, file_modication_datetime=None):
        plugin_manager.register(self.listener)
        plugin_manager.register(self.logger)

        for test_case in structured_data["tests"]:
            self.process_test_case(test_case, file_modication_datetime=file_modication_datetime)

        plugin_manager.unregister(plugin=self.listener)
        plugin_manager.unregister(plugin=self.logger)

    def process_test_case(self, test_case, file_modication_datetime=None):

        with self.lifecycle.schedule_test_case() as test_result:
            test_index = test_case["id"]
            test_data = test_case.get("data") or {}
            job = test_data.get("job") or {}
            test_result.name = test_index
            self._record_start_stop(test_result, file_modication_datetime, job, test_data)

            test_result.fullName = test_index
            test_result.testCaseId = md5(test_index)
            # Maybe an option to swap this - seems like it is causing all runs to seem like
            # retries instead of history proper?
            # test_result.historyId = md5(test_index)
            test_result.historyId = md5(str(uuid.uuid4()))
            tool_id = self._record_suite_labels(test_result, test_index, test_data, job)

            self._attach_data(
                "test_data", json.dumps(test_data, indent=JSON_INDENT), attachment_type=AttachmentType.JSON
            )
            for key in ["stderr", "stdout", "command_line", "external_id", "job_messages"]:
                val = job.get(key)
                if not val:
                    continue
                if isinstance(val, list):
                    attachment_type = AttachmentType.JSON
                    # job messages
                    val = json.dumps(val, indent=JSON_INDENT)
                else:
                    if not val.strip():
                        continue
                    attachment_type = AttachmentType.TEXT
                self._attach_data(key, val, attachment_type=attachment_type)

            problem_message = None
            for key in ["execution_problem", "output_problems"]:
                val = test_data.get(key)
                if not val:
                    continue
                if isinstance(val, list) and val:
                    # remove duplicated messages...
                    val = list(set(val))

                    attachment_type = AttachmentType.HTML
                    as_html_list = "<ul>"
                    as_html_list += "\n".join([f"<li><pre>{v}</pre></li>" for v in val])
                    as_html_list += "</ul>"
                    problem_message = val[0]
                    val = as_html_list
                else:
                    if not val.strip():
                        continue
                    attachment_type = AttachmentType.TEXT
                    problem_message = val
                self._attach_data(key, val, attachment_type=attachment_type)

            if problem_message is None and "job_messages" in job:
                job_messages = job.get("job_messages")
                if job_messages:
                    problem_message = str(job_messages)

            test_result.labels.append(Label(name=LabelType.FRAMEWORK, value="planemo"))
            test_result.labels.append(Label(name=LabelType.LANGUAGE, value=platform_label()))

            self._record_tool_link(test_result, tool_id)
            self._record_status(test_result, test_data)
            if test_result.status in [Status.BROKEN, Status.FAILED]:
                test_result.statusDetails = StatusDetails(
                    message=escape_non_unicode_symbols(problem_message or "Unknown problem"), trace=None
                )

        self.lifecycle.write_test_case()

    def _record_start_stop(self, test_result, file_modication_datetime, job, test_data):
        start_datetime = file_modication_datetime
        end_datetime = file_modication_datetime
        if "start_datetime" in test_data:
            start_datetime = parser.parse(test_data["start_datetime"])
        elif "create_time" in job:
            start_datetime = parser.parse(job["create_time"])

        if "end_datetime" in test_data:
            end_datetime = parser.parse(test_data["end_datetime"])
        if "update_time" in job:
            end_datetime = parser.parse(job["update_time"])

        if start_datetime is not None:
            test_result.start = int(round(start_datetime.timestamp() * 1000))
        if end_datetime is not None:
            test_result.stop = int(round(end_datetime.timestamp() * 1000))

    def _record_suite_labels(self, test_result, test_index, test_data, job):
        tool_id = None
        index_match = WORKFLOW_INDEX_MATCH.match(test_index)
        if "tool_id" in test_data:
            tool_id = test_data["tool_id"]
            test_result.labels.append(Label(name=LabelType.PARENT_SUITE, value=tool_id))
        elif "tool_id" in job:
            tool_id = job["tool_id"]
            test_result.labels.append(Label(name=LabelType.PARENT_SUITE, value=tool_id))
        elif index_match:
            test_result.labels.append(Label(name=LabelType.PARENT_SUITE, value=index_match.group(1)))

        if "tool_version" in test_data:
            test_result.labels.append(Label(name=LabelType.SUITE, value=test_data["tool_version"]))
        elif "tool_version" in job:
            test_result.labels.append(Label(name=LabelType.SUITE, value=job["tool_version"]))

        if "test_index" in test_data:
            test_result.labels.append(Label(name=LabelType.SUB_SUITE, value=str(test_data["test_index"])))
        elif index_match:
            test_result.labels.append(Label(name=LabelType.SUB_SUITE, value=index_match.group(2)))

        return tool_id

    def _record_tool_link(self, test_result, tool_id):
        if tool_id and "repos" in tool_id:
            tool_parts = tool_id.split("/")
            if len(tool_parts) >= 4:
                link = Link(LinkType.LINK, "https://%s" % "/".join(tool_parts[0:4]), "Tool Repository")
                test_result.links.append(link)

    def _record_status(self, test_result, test_data):
        status = test_data.get("status", "error")
        if status == "success":
            test_result.status = Status.PASSED
        elif status == "failure":
            test_result.status = Status.FAILED
        elif status == "skip":
            test_result.status = Status.SKIPPED
        else:
            test_result.status = Status.BROKEN

    def _attach_data(self, key, val, attachment_type=AttachmentType.TEXT):
        self.lifecycle.attach_data(uuid4(), val, name=key, attachment_type=attachment_type, extension=None)


def write_results(results_path, structured_data, **kwds):
    AllureWriter(results_path).process(structured_data)



"""Define the context around Planemo computation.

Abstractions for cross cutting concerns (logging, workspace management,
etc.).
"""
import abc
import logging.config
import os
import shutil
import sys
import traceback
from typing import (
    Any,
    Dict,
    Optional,
    TYPE_CHECKING,
)
from urllib.request import urlopen

from planemo.config import read_global_config

if TYPE_CHECKING:
    from planemo.config import OptionSource


class PlanemoContextInterface(metaclass=abc.ABCMeta):
    """Interface Planemo operations use to access workspace context."""

    @abc.abstractmethod
    def set_option_source(self, param_name, option_source, force=False):
        """Specify how an option was set."""

    @abc.abstractmethod
    def get_option_source(self, param_name):
        """Return OptionSource value indicating how the option was set."""

    @abc.abstractproperty
    def global_config(self):
        """Read Planemo's global configuration."""

    @abc.abstractmethod
    def log(self, msg, *args):
        """Log a message."""

    @abc.abstractmethod
    def vlog(self, msg, *args, **kwds):
        """Log a message only if verbose is enabled."""

    @abc.abstractproperty
    def workspace(self):
        """Create and return Planemo's workspace."""

    @abc.abstractproperty
    def galaxy_profiles_directory(self):
        """Create a return a directory for storing Galaxy profiles."""

    @abc.abstractmethod
    def cache_download(self, url, destination):
        """Use workspace to cache download of ``url``."""


class PlanemoContext(PlanemoContextInterface):
    """Implementation of ``PlanemoContextInterface``"""

    planemo_directory: Optional[str]

    def __init__(self) -> None:
        """Construct a Context object using execution environment."""
        self.home = os.getcwd()
        self._global_config: Optional[Dict] = None
        # Will be set by planemo CLI driver
        self.verbose = False
        self.planemo_config: Optional[str] = None
        self.planemo_directory = None
        self.option_source: Dict[str, "OptionSource"] = {}

    def set_option_source(self, param_name: str, option_source: "OptionSource", force: bool = False) -> None:
        """Specify how an option was set."""
        if not force:
            assert param_name not in self.option_source, f"Option source for [{param_name}] already set"
        self.option_source[param_name] = option_source

    def get_option_source(self, param_name: str) -> "OptionSource":
        """Return OptionSource value indicating how the option was set."""
        assert param_name in self.option_source, f"No option source for [{param_name}]"
        return self.option_source[param_name]

    @property
    def global_config(self) -> Dict[str, Any]:
        """Read Planemo's global configuration.

        As defined most simply by ~/.planemo.yml.
        """
        if self._global_config is None:
            self._global_config = read_global_config(self.planemo_config)
        return self._global_config or {}

    def log(self, msg: str, *args) -> None:
        """Log a message."""
        if args:
            msg %= args
        self._log_message(msg)

    def vlog(self, msg: str, *args, **kwds) -> None:
        """Log a message only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args)
            if kwds.get("exception", False):
                traceback.print_exc(file=sys.stderr)

    @property
    def workspace(self) -> str:
        """Create and return Planemo's workspace.

        By default this will be ``~/.planemo``.
        """
        if not self.planemo_directory:
            raise Exception("No planemo workspace defined.")
        workspace = self.planemo_directory
        return self._ensure_directory(workspace, "workspace")

    @property
    def galaxy_profiles_directory(self) -> str:
        """Create a return a directory for storing Galaxy profiles."""
        path = os.path.join(self.workspace, "profiles")
        return self._ensure_directory(path, "Galaxy profiles")

    def cache_download(self, url, destination):
        """Use workspace to cache download of ``url``."""
        cache = os.path.join(self.workspace, "cache")
        if not os.path.exists(cache):
            os.makedirs(cache)
        filename = os.path.basename(url)
        cache_destination = os.path.join(cache, filename)
        if not os.path.exists(cache_destination):
            with urlopen(url) as fh:
                content = fh.read()
            if len(content) == 0:
                raise Exception(f"Failed to download [{url}].")
            with open(cache_destination, "wb") as f:
                f.write(content)

        shutil.copy(cache_destination, destination)

    def _ensure_directory(self, path: str, name: str) -> str:
        if not os.path.exists(path):
            os.makedirs(path)
        if not os.path.isdir(path):
            raise Exception(f"Planemo {name} directory [{path}] unavailable.")
        return path

    def _log_message(self, message):
        """Extension point for overriding I/O."""
        print(message)


def configure_standard_planemo_logging(verbose: bool) -> None:
    """Configure Planemo's default logging rules."""
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "verbose": {"format": "%(name)s %(levelname)s %(asctime)s: %(message)s"},
            "simple": {"format": "%(name)s %(levelname)s: %(message)s"},
        },
        "handlers": {
            "console": {
                "level": "DEBUG",
                "class": "logging.StreamHandler",
                "formatter": "simple" if not verbose else "verbose",
            },
        },
        "loggers": {
            # Suppress CWL is beta warning, for Planemo purposes - it is absolutely not.
            "galaxy.tools.parser.factory": {
                "handlers": ["console"],
                "propagate": False,
                "level": "ERROR" if not verbose else "DEBUG",
            },
            "galaxy.tools.deps.commands": {
                "handlers": ["console"],
                "propagate": False,
                "level": "ERROR" if not verbose else "DEBUG",
            },
            "galaxy": {
                "handlers": ["console"],
                "propagate": False,
                "level": "INFO" if not verbose else "DEBUG",
            },
            # @jmchilton
            # I'm fixing up Planemo's lint functionality for CWL and I keep seeing this for the
            # schema metadata stuff (e.g. in the workflows repo). "rdflib.term WARNING:
            # http://schema.org/docs/!DOCTYPE html does not look like a valid URI, trying to
            # serialize this will break.". I'm going to suppress this warning I think, or are the
            # examples wrong and should declare their namespaces differently in some way?
            # @mr-c
            # That particular warning is worth suppressing. A PR to silence it permanently would be very welcome!
            # https://github.com/RDFLib/rdflib/blob/master/rdflib/term.py#L225
            "rdflib.term": {
                "handlers": ["console"],
                "propagate": False,
                "level": "ERROR" if not verbose else "DEBUG",
            },
        },
        "root": {
            "handlers": ["console"],
            "propagate": False,
            "level": "WARNING" if not verbose else "DEBUG",
        },
    }
    logging.config.dictConfig(logging_config)


__all__ = (
    "configure_standard_planemo_logging",
    "PlanemoContextInterface",
    "PlanemoContext",
)



"""Planemo-specific wrappers around galaxy-tool-util tool functionality."""

import os
import sys
import traceback
from typing import (
    Iterable,
    Iterator,
    List,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from galaxy.tool_util import loader_directory
from galaxy.tool_util.fetcher import ToolLocationFetcher
from galaxy.tool_util.loader_directory import is_tool_load_error
from galaxy.tool_util.parser.interface import ToolSource

from planemo.io import (
    error,
    info,
)

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext

SKIP_XML_MESSAGE = "Skipping XML file - does not appear to be a tool %s."
SHED_FILES = ["tool_dependencies.xml", "repository_dependencies.xml"]
LOAD_ERROR_MESSAGE = "Error loading tool with path %s"


def uri_to_path(ctx: "PlanemoCliContext", uri: str) -> str:
    """Fetch URI to a local path."""
    fetcher = ToolLocationFetcher()
    return fetcher.to_tool_path(uri)


def uris_to_paths(ctx, uris):
    """Fetch multiple URIs to a local path."""
    fetcher = ToolLocationFetcher()
    paths = []
    for uri in uris:
        path = fetcher.to_tool_path(uri)
        paths.append(path)
    return paths


def yield_tool_sources_on_paths(
    ctx: "PlanemoCliContext",
    paths: Iterable[str],
    recursive: bool = False,
    yield_load_errors: bool = True,
    exclude_deprecated: bool = False,
) -> Iterator[Tuple[str, Union[ToolSource, object]]]:
    """Walk paths and yield ToolSource objects discovered."""
    for path in paths:
        for (tool_path, tool_source) in yield_tool_sources(ctx, path, recursive, yield_load_errors):
            if exclude_deprecated and "deprecated" in tool_path:
                continue
            yield (tool_path, tool_source)


def yield_tool_sources(
    ctx: "PlanemoCliContext", path: str, recursive: bool = False, yield_load_errors: bool = True
) -> Iterator[Tuple[str, Union[ToolSource, object]]]:
    """Walk single path and yield ToolSource objects discovered."""
    tools = load_tool_sources_from_path(
        path,
        recursive,
        register_load_errors=True,
    )
    for (tool_path, tool_source) in tools:
        if is_tool_load_error(tool_source):
            if yield_load_errors:
                yield (tool_path, tool_source)
            else:
                error(LOAD_ERROR_MESSAGE % tool_path)
            continue

        if not _is_tool_source(ctx, tool_path, tool_source):
            continue
        yield (tool_path, tool_source)


def load_tool_sources_from_path(
    path: str, recursive: bool, register_load_errors: bool = False
) -> List[Tuple[str, Union[ToolSource, object]]]:
    """Generate a list for tool sources found down specified path."""
    return loader_directory.load_tool_sources_from_path(
        path,
        _load_exception_handler,
        recursive=recursive,
        register_load_errors=register_load_errors,
    )


def _load_exception_handler(path, exc_info):
    error(LOAD_ERROR_MESSAGE % path)
    traceback.print_exception(*exc_info, limit=1, file=sys.stderr)


def _is_tool_source(ctx: "PlanemoCliContext", tool_path: str, tool_source: "ToolSource") -> bool:
    if os.path.basename(tool_path) in SHED_FILES:
        return False
    root = getattr(tool_source, "root", None)
    if root is not None:
        if root.tag != "tool":
            if ctx.verbose:
                info(SKIP_XML_MESSAGE % tool_path)
            return False
    return True


__all__ = (
    "is_tool_load_error",
    "load_tool_sources_from_path",
    "yield_tool_sources",
    "yield_tool_sources_on_paths",
)



"""Ensure requirements are matched in best practice conda channels."""

from planemo.conda import (
    BEST_PRACTICE_CHANNELS,
    best_practice_search,
    tool_source_conda_targets,
)

lint_tool_types = ["*"]


def lint_requirements_in_conda(tool_source, lint_ctx):
    """Check requirements of tool source against best practice Conda channels."""
    conda_targets = tool_source_conda_targets(tool_source)
    if not conda_targets:
        lint_ctx.warn("No valid package requirement tags found to check against Conda.")
        return

    for conda_target in conda_targets:
        (best_hit, exact) = best_practice_search(conda_target)
        conda_target_str = conda_target.package
        if conda_target.version:
            conda_target_str += "@%s" % (conda_target.version)
        if best_hit and exact:
            template = "Requirement [%s] matches target in best practice Conda channel [%s]."
            message = template % (conda_target_str, best_hit.get("channel"))
            lint_ctx.info(message)
        elif best_hit:
            template = (
                "Requirement [%s] doesn't exactly match available version [%s] in best practice Conda channel [%s]."
            )
            message = template % (conda_target_str, best_hit["version"], best_hit.get("channel"))
            lint_ctx.warn(message)
        else:
            template = "Requirement [%s] doesn't match any recipe in a best practice conda channel [%s]."
            message = template % (conda_target_str, BEST_PRACTICE_CHANNELS)
            lint_ctx.warn(message)



"""Module describing the planemo ``docker_build`` command."""

import click
from galaxy.tool_util.deps import dockerfiles

from planemo import options
from planemo.cli import command_function
from planemo.io import error


@click.command("docker_build")
@options.optional_tools_arg()
@click.option("--dockerfile", default=None)
@click.option("--docker_image_cache", default=None)
@options.docker_cmd_option()
@options.docker_sudo_option()
@options.docker_sudo_cmd_option()
@options.docker_host_option()
@command_function
def cli(ctx, path=".", dockerfile=None, **kwds):
    """Build (and optionally cache) Docker images.

    Loads the tool or tools referenced by ``TOOL_PATH`` (by default all tools
    in current directory), and ensures they all reference the same Docker image
    and then attempts to find a Dockerfile for these tools (can be explicitly
    specified with ``--dockerfile`` but by default it will check the tool's
    directory and the current directory as well).

    This command will then build and tag the image so it is ready to be tested
    and published. The docker_shell command be used to test out the built
    image.

    \b
        % planemo docker_build bowtie2.xml # asssumes Dockerfile in same dir
        % planemo docker_shell --from_tag bowtie2.xml

    This can optionally also cache the images.
    """
    dockerfiles.dockerfile_build(path=path, dockerfile=dockerfile, error=error, **kwds)



"""Module describing the planemo ``mulled_init`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.mulled import build_involucro_context


@click.command("mulled_init")
@options.mulled_options()
@command_function
def cli(ctx, **kwds):
    """Download and install involucro for mull command.

    This will happen automatically when using the mull command, but this can
    be pre-installed in an environment using this command.
    """
    build_involucro_context(ctx, **kwds)



"""Planemo specific utilities for dealing with conda.

The extend galaxy-tool-util's features with planemo specific idioms.
"""


import collections
import os
import threading
from copy import deepcopy
from typing import TYPE_CHECKING

from galaxy.tool_util.deps import conda_util
from galaxy.tool_util.deps.conda_util import CondaContext
from galaxy.util import unicodify

from planemo.exit_codes import (
    EXIT_CODE_FAILED_DEPENDENCIES,
    ExitCodeException,
)
from planemo.io import (
    error,
    shell,
)
from planemo.tools import yield_tool_sources_on_paths

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext

MESSAGE_ERROR_FAILED_INSTALL = "Attempted to install conda and failed."
MESSAGE_ERROR_CANNOT_INSTALL = "Cannot install Conda - perhaps due to a failed installation or permission problems."
MESSAGE_ERROR_NOT_INSTALLING = (
    "Conda not configured - run ``planemo conda_init`` or pass ``--conda_auto_init`` to continue."
)

BEST_PRACTICE_CHANNELS = ["conda-forge", "bioconda", "defaults"]


def build_conda_context(ctx: "PlanemoCliContext", **kwds) -> CondaContext:
    """Build a galaxy-tool-util CondaContext tailored to planemo use.

    Using planemo's common command-line/global config options.
    """
    condarc_override_default = os.path.join(ctx.workspace, "condarc")
    conda_prefix = kwds.get("conda_prefix", None)
    use_planemo_shell = kwds.get("use_planemo_shell_exec", True)
    ensure_channels = kwds.get("conda_ensure_channels", "")
    condarc_override = kwds.get("condarc", condarc_override_default)
    use_local = kwds.get("conda_use_local", False)
    shell_exec = shell if use_planemo_shell else None
    conda_context = conda_util.CondaContext(
        conda_prefix=conda_prefix,
        ensure_channels=ensure_channels,
        condarc_override=condarc_override,
        use_local=use_local,
        shell_exec=shell_exec,
    )
    handle_auto_init = kwds.get("handle_auto_init", False)
    if handle_auto_init and not conda_context.is_installed():
        auto_init = kwds.get("conda_auto_init", True)
        failed = True
        if auto_init:
            if conda_context.can_install_conda():
                if conda_util.install_conda(conda_context):
                    error(MESSAGE_ERROR_FAILED_INSTALL)
                else:
                    failed = False
            else:
                error(MESSAGE_ERROR_CANNOT_INSTALL)
        else:
            error(MESSAGE_ERROR_NOT_INSTALLING)

        if failed:
            raise ExitCodeException(EXIT_CODE_FAILED_DEPENDENCIES)
    if handle_auto_init:
        conda_context.ensure_conda_build_installed_if_needed()
    return conda_context


def collect_conda_targets(ctx, paths, recursive=False, found_tool_callback=None):
    """Load CondaTarget objects from supplied artifact sources.

    If a tool contains more than one requirement, the requirements will each
    appear once in the output.
    """
    conda_targets = set()
    real_paths = []
    for path in paths:
        if not os.path.exists(path):
            targets = target_str_to_targets(path)
            [conda_targets.add(_) for _ in targets]
        else:
            real_paths.append(path)

    for (tool_path, tool_source) in yield_tool_sources_on_paths(
        ctx, real_paths, recursive=recursive, exclude_deprecated=True
    ):
        if found_tool_callback:
            found_tool_callback(tool_path)
        for target in tool_source_conda_targets(tool_source):
            conda_targets.add(target)
    return conda_targets


# Copied and modified from mulled stuff - need to syncronize these concepts.
def target_str_to_targets(targets_raw):
    def parse_target(target_str):
        if "=" in target_str:
            package_name, version = target_str.split("=", 1)
        else:
            package_name = target_str
            version = None
        target = conda_util.CondaTarget(package_name, version)
        return target

    targets = [parse_target(_) for _ in targets_raw.split(",")]
    return targets


def collect_conda_target_lists(ctx, paths, recursive=False, found_tool_callback=None):
    """Load CondaTarget lists from supplied artifact sources.

    If a tool contains more than one requirement, the requirements will all
    appear together as one list element of the output list.
    """
    conda_target_lists, _ = collect_conda_target_lists_and_tool_paths(
        ctx, paths, recursive=recursive, found_tool_callback=found_tool_callback
    )
    return conda_target_lists


def collect_conda_target_lists_and_tool_paths(ctx, paths, recursive=False, found_tool_callback=None):
    """Load CondaTarget lists from supplied artifact sources.

    If a tool contains more than one requirement, the requirements will all
    appear together as one list element of the output list.
    """
    conda_target_lists = set()
    tool_paths = collections.defaultdict(list)
    for (tool_path, tool_source) in yield_tool_sources_on_paths(
        ctx, paths, recursive=recursive, yield_load_errors=False
    ):
        try:
            if found_tool_callback:
                found_tool_callback(tool_path)
            targets = frozenset(tool_source_conda_targets(tool_source))
            conda_target_lists.add(targets)
            tool_paths[targets].append(tool_path)
        except Exception as e:
            ctx.log(f"Error while collecting list of conda targets for '{tool_path}': {unicodify(e)}")

    # Turn them into lists so the order matches before returning...
    conda_target_lists = list(conda_target_lists)
    conda_target_tool_paths = [tool_paths[c] for c in conda_target_lists]

    return conda_target_lists, conda_target_tool_paths


def tool_source_conda_targets(tool_source):
    """Load CondaTarget object from supplied abstract tool source."""
    requirements, _ = tool_source.parse_requirements_and_containers()
    return conda_util.requirements_to_conda_targets(requirements)


best_practice_search_first = threading.local()


def best_practice_search(conda_target, conda_context=None, platform=None):
    # Call it in offline mode after the first time.
    try:
        best_practice_search_first.previously_called
        # TODO: Undo this...
        offline = False
    except AttributeError:
        best_practice_search_first.previously_called = True
        offline = False

    if conda_context:
        if conda_context.ensure_channels != BEST_PRACTICE_CHANNELS:
            conda_context = deepcopy(conda_context)
            conda_context.ensure_channels = BEST_PRACTICE_CHANNELS
    else:
        conda_context = conda_util.CondaContext(ensure_channels=BEST_PRACTICE_CHANNELS)
    return conda_util.best_search_result(
        conda_target,
        conda_context=conda_context,
        offline=offline,
        platform=platform,
    )


__all__ = (
    "BEST_PRACTICE_CHANNELS",
    "best_practice_search",
    "build_conda_context",
    "collect_conda_targets",
    "collect_conda_target_lists",
    "collect_conda_target_lists_and_tool_paths",
    "tool_source_conda_targets",
)



"""Module describing the planemo ``shed_update`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function
from planemo.io import (
    captured_io_for_xunit,
    error,
    info,
)
from planemo.reports.xunit_handler import handle_report_xunit_kwd


@click.command("shed_update")
@options.report_xunit()
@options.shed_publish_options()
@options.shed_upload_options()
@options.shed_skip_upload()
@options.shed_skip_metadata()
@command_function
def cli(ctx, paths, **kwds):
    """Update Tool Shed repository.

    By default this command will update both repository metadata
    from ``.shed.yml`` and upload new contents from the repository
    directory.

    \b
        % planemo shed_update

    This will update the main tool shed with the repository defined
    by a ``.shed.yml`` file in the current working directory. Both
    the location of the ``.shed.yml`` and the tool shed to upload to
    can be easily configured. For instance, the following command can
    be used if ``.shed.yml`` if contained in ``path/to/repo`` and the
    desire is to update the test tool shed.

    \b
        % planemo shed_update --shed_target testtoolshed path/to/repo

    Another important option is ``--check_diff`` - this doesn't affect the
    updating of shed metadata but it will check for differences before
    uploading new contents to the tool shed. This may important because the
    tool shed will automatically populate certain attributes in tool shed
    artifact files (such as ``tool_dependencies.xml``) and this may
    cause unwanted installable revisions to be created when there are no
    important changes.

    The lower-level ``shed_upload`` command should be used instead if
    the repository doesn't define complete metadata in a ``.shed.yml``.
    """
    # In a little bit of cheating, we're defining this variable here to collect
    # a "report" on the shed_update command
    collected_data = {
        "results": {
            "total": 0,
            "errors": 0,
            "failures": 0,
            "skips": 0,
        },
        "suitename": "update",
        "tests": [],
    }

    shed_context = shed.get_shed_context(ctx, **kwds)

    def update(realized_repository):
        collected_data["results"]["total"] += 1
        skip_upload = kwds["skip_upload"]
        skip_metadata = kwds["skip_metadata"]
        upload_ret_code = 0
        upload_ok = True

        captured_io = {}
        if not skip_upload:
            with captured_io_for_xunit(kwds, captured_io):
                upload_ret_code = shed.upload_repository(ctx, realized_repository, **kwds)
                upload_ok = not upload_ret_code

        repo_result = {
            "classname": realized_repository.name,
            "time": captured_io.get("time", None),
            "name": "shed-update",
            "stdout": captured_io.get("stdout", None),
            "stderr": captured_io.get("stderr", None),
        }

        # Now that we've uploaded (or skipped appropriately), collect results.
        if upload_ret_code == 2:
            collected_data["results"]["failures"] += 1
            message = f"Failed to update repository '{realized_repository.name}' as it does not exist on the {shed_context.label}."
            repo_result.update(
                {
                    "errorType": "FailedUpdate",
                    "errorMessage": message,
                }
            )
            collected_data["tests"].append(repo_result)
            error(message)
            return upload_ret_code

        exit = 0
        metadata_ok = True
        repository_destination_label = f"repository '{realized_repository.name}' on the {shed_context.label}"
        if not skip_metadata:
            repo_id = shed.handle_force_create(realized_repository, ctx, shed_context, **kwds)
            # failing to create the repo, give up
            if repo_id is None:
                exit = shed.report_non_existent_repository(realized_repository)
                metadata_ok = False
                error("Failed to update metadata for %s." % repository_destination_label)
            else:
                metadata_ok = realized_repository.update(ctx, shed_context, repo_id)
                if metadata_ok:
                    info("Repository metadata updated successfully for %s." % repository_destination_label)
        else:
            info("Skipping metadata update for %s" % repository_destination_label)

        if metadata_ok and upload_ok:
            pass
        elif upload_ok:
            collected_data["results"]["skips"] += 1
            repo_result.update(
                {
                    "errorType": "FailedMetadata",
                    "errorMessage": "Failed to update repository metadata",
                }
            )
            if not skip_upload:
                error(
                    "Repository contents updated but failed to update metadata for %s." % repository_destination_label
                )
            exit = exit or 1
        else:
            collected_data["results"]["failures"] += 1
            repo_result.update(
                {
                    "errorType": "FailedUpdate",
                    "errorMessage": "Failed to update repository",
                }
            )
            if metadata_ok:
                error("Failed to update repository contents for %s." % repository_destination_label)
            else:
                error("Failed to update repository contents and metadata for %s." % repository_destination_label)
            exit = exit or 1
        collected_data["tests"].append(repo_result)
        return exit

    exit_code = shed.for_each_repository(ctx, update, paths, **kwds)

    handle_report_xunit_kwd(kwds, collected_data)

    sys.exit(exit_code)



"""Module describing the planemo ``conda_install`` command."""
import click
from galaxy.tool_util.deps import conda_util

from planemo import options
from planemo.cli import command_function
from planemo.conda import (
    build_conda_context,
    collect_conda_targets,
)
from planemo.io import coalesce_return_codes


@click.command("conda_install")
@options.optional_tools_or_packages_arg(multiple=True)
@options.recursive_option()
@options.conda_target_options()
@options.conda_global_option()
@options.conda_auto_init_option()
@command_function
def cli(ctx, paths, **kwds):
    """Install conda packages for tool requirements."""
    conda_context = build_conda_context(ctx, handle_auto_init=True, **kwds)
    return_codes = []
    for conda_target in collect_conda_targets(ctx, paths, recursive=kwds["recursive"]):
        ctx.log("Install conda target %s" % conda_target)
        return_code = conda_util.install_conda_target(
            conda_target, conda_context=conda_context, skip_environment=kwds.get("global", False)
        )
        return_codes.append(return_code)
    return coalesce_return_codes(return_codes, assert_at_least_one=True)



"""Module defines a planemo abstraction around running cwltool.

cwltool is an executable Python script and library mostly maintained by
Peter Amstutz and serves the reference implementation for the CWL.
It can be found at https://github.com/common-workflow-language/cwltool,
"""
import json
import tempfile
from typing import (
    Any,
    Dict,
    Optional,
    TYPE_CHECKING,
    Union,
)

from galaxy.tool_util.cwl.cwltool_deps import (
    ensure_cwltool_available,
    main,
)

from planemo.deps import ensure_dependency_resolvers_conf_configured
from planemo.io import (
    error,
    real_io,
)
from planemo.runnable import (
    ErrorRunResponse,
    Runnable,
    SuccessfulRunResponse,
)

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext

JSON_PARSE_ERROR_MESSAGE = "Failed to parse JSON from cwltool output [%s] " "in file [%s]. cwltool logs [%s]."


class CwlToolRunResponse(SuccessfulRunResponse):
    """Describe the resut of a cwltool invocation."""

    def __init__(
        self,
        runnable: "Runnable",
        log: str,
        outputs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(runnable=runnable)
        self._log = log
        self._outputs = outputs

    @property
    def log(self) -> str:
        return self._log

    @property
    def job_info(self) -> None:
        return None

    @property
    def invocation_details(self) -> None:
        return None

    @property
    def outputs_dict(self) -> Optional[Dict[str, Any]]:
        return self._outputs


def run_cwltool(
    ctx: "PlanemoCliContext", runnable: "Runnable", job_path: str, **kwds
) -> Union[ErrorRunResponse, CwlToolRunResponse]:
    """Translate planemo kwds to cwltool kwds and run cwltool main function."""
    ensure_cwltool_available()

    args = []
    if ctx.verbose:
        args.append("--verbose")
    output_directory = kwds.get("output_directory", None)
    if output_directory:
        args.append("--outdir")
        args.append(output_directory)
    if kwds.get("no_container", False):
        args.append("--no-container")
        ensure_dependency_resolvers_conf_configured(ctx, kwds)
        args.append("--beta-dependency-resolvers-configuration")
        args.append(kwds["dependency_resolvers_config_file"])
    if kwds.get("mulled_containers"):
        args.append("--beta-use-biocontainers")

    if kwds.get("non_strict_cwl", False):
        args.append("--non-strict")

    args.extend([runnable.path, job_path])
    ctx.vlog("Calling cwltool with arguments %s" % args)
    with tempfile.NamedTemporaryFile("w") as tmp_stdout, tempfile.NamedTemporaryFile("w") as tmp_stderr:
        # cwltool passes sys.stderr to subprocess.Popen - ensure it has
        # and actual fileno.
        with real_io():
            ret_code = main.main(
                args,
                stdout=tmp_stdout,
                stderr=tmp_stderr,
            )
        tmp_stdout.flush()
        tmp_stderr.flush()
        with open(tmp_stderr.name) as stderr_f:
            log = stderr_f.read()
            ctx.vlog("cwltool log output [%s]" % log)
        with open(tmp_stdout.name) as stdout_f:
            try:
                result = json.load(stdout_f)
            except ValueError:
                message = JSON_PARSE_ERROR_MESSAGE % (open(tmp_stdout.name).read(), tmp_stdout.name, log)
                error(message)
                raise Exception(message)

        if ret_code != 0:
            return ErrorRunResponse("Error running cwltool", log=log)
        outputs = result
    return CwlToolRunResponse(
        runnable,
        log,
        outputs=outputs,
    )


__all__ = ("run_cwltool",)
